from dataclasses import dataclass
from .....utils import DictUtils
import os

LOGIN_SESSION_COLLECTION = os.environ.get('LOGIN_SESSION_COLLECTION', 'login_sessions')


@dataclass
class LoginSessionFields:
    id = "id"
    user_uid = "user_uid"
    session_token = "session_token"
    device_name = "device_name"
    device_type = "device_type"
    ip_address = "ip_address"
    user_agent = "user_agent"
    is_active = "is_active"
    mfa_verified = "mfa_verified"
    created_at = "created_at"
    last_seen_at = "last_seen_at"
    revoked_at = "revoked_at"

    @staticmethod
    def keys():
        return DictUtils.get_keys(LoginSessionFieldsProps)

    @staticmethod
    def filtered_keys(field, condition=True):
        mutable = DictUtils.filter(LoginSessionFieldsProps, DictUtils.get_keys(LoginSessionFieldsProps), field, condition)
        return DictUtils.get_keys(mutable)


LoginSessionFieldsProps = {
    LoginSessionFields.id: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    LoginSessionFields.user_uid: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    LoginSessionFields.session_token: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    LoginSessionFields.device_name: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    LoginSessionFields.device_type: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    LoginSessionFields.ip_address: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    LoginSessionFields.user_agent: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    LoginSessionFields.is_active: {
        "type": bool,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": True,
        "pickable": True
    },
    LoginSessionFields.mfa_verified: {
        "type": bool,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": False,
        "pickable": True
    },
    LoginSessionFields.created_at: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    LoginSessionFields.last_seen_at: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    LoginSessionFields.revoked_at: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
}

STANDARD_FIELDS = LoginSessionFields.filtered_keys('pickable', True)



from dataclasses import dataclass 
from ...models.adgroups.base_adgroup_fields import BaseAdGroupFields, BaseAdGroupFieldsProps
import os

DISPLAY_ADGROUPS_COLLECTION = os.environ.get('GADS_ADGROUPS_DISPLAY_COLLECTION');

@dataclass
class AdGroupDisplayFields(BaseAdGroupFields):
    targetedPlacements = "targetedPlacements"
    excludedPlacements = "excludedPlacements"
    websites = "websites"
    youtubeChannels = "youtubeChannels"
    youtubeVideos = "youtubeVideos"
    images = "images"
    imagesNative = "imagesNative"
    user_interest = "user_interest"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.class_fields_props = AdGroupDisplayFieldsProps



AdGroupDisplayFieldsProps = {
    **BaseAdGroupFieldsProps,
    AdGroupDisplayFields.targetedPlacements: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    AdGroupDisplayFields.excludedPlacements: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    AdGroupDisplayFields.websites: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    AdGroupDisplayFields.youtubeChannels: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    AdGroupDisplayFields.youtubeVideos: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    AdGroupDisplayFields.user_interest: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    AdGroupDisplayFields.images: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    AdGroupDisplayFields.imagesNative: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    }
}


STANDARD_FIELDS = AdGroupDisplayFields(fields_props=AdGroupDisplayFieldsProps).filtered_keys('pickable', True)
ADGROUP_DISPLAY_PICKABLE_FIELDS = AdGroupDisplayFields(fields_props=AdGroupDisplayFieldsProps).filtered_keys('pickable', True)




import os
from sendgrid.helpers.mail import Email, Personalization
from sendgrid import SendGridAPIClient, Mail as SendGridMail
import urllib.parse

from datetime import datetime
from adafri.v1.user import User
from adafri.utils import generate_random_code, ResponseStatus, ApiResponse, Error, StatusCode
from adafri.v1.auth.oauth import Code
from adafri.utils import (Object, boolean, Crypto)

VALIDATION_FIELD = '_emailValidationSendDate'
PSW_RESET_FIELD = '_pwResetSendDate'
STATUS_FIELD = 'status'

SENDGRID_API_KEY = os.environ.get('SENDGRID_API_KEY');
REGISTRATION_TEMPLATE_ID = os.environ.get('REGISTRATION_TEMPLATE_ID')
PAYMENT_TEMPLATE_ID = os.environ.get('PAYMENT_TEMPLATE_ID')
EMAIL_VERIFICATION_TEMPLATE_ID = os.environ.get('EMAIL_VERIFICATION_TEMPLATE_ID')
try:
    os_value = os.environ.get('DEFAULT_EMAIL_TIME_TO_RESEND');
    if os_value is None:
        DEFAULT_EMAIL_TIME_TO_RESEND = 10
    else:
        DEFAULT_EMAIL_TIME_TO_RESEND = int(str(os_value))
except:
    DEFAULT_EMAIL_TIME_TO_RESEND = 10;



def compareTimes(date_string: str):
    date_format = "%Y-%m-%d, %H:%M:%S"
    date = datetime.strptime(date_string, date_format);
    now = datetime.now();
    difference = now - date;
    days, seconds = difference.days, difference.seconds
    hours = days * 24 + seconds // 3600
    minutes = (seconds % 3600) // 60
    seconds = seconds % 60
    if minutes >= DEFAULT_EMAIL_TIME_TO_RESEND:
        return {"status": "ok", "difference": minutes};
    return {"status": "error", "difference": minutes};


def setTimeSended(user: User, field=VALIDATION_FIELD):
    try:
        date_format = "%Y-%m-%d, %H:%M:%S"
        now = datetime.now()
        date_obj = datetime.strftime(now, date_format)
        update_value = {};
        update_value[field] = date_obj
        update = user.document_reference().set(update_value, merge=True)
        return update_value
    except:
        return None;

def getTimeSended(user: User, field=VALIDATION_FIELD):
    statement = {};
    if user is None:
        statement[STATUS_FIELD]='error'
        statement[field]=None
        return statement
    validation = getattr(user, field, None)
    if validation is not None and bool(validation):
        statement[STATUS_FIELD]='ok'
        statement[field]=validation
        return statement
    else:
        statement[STATUS_FIELD]='ok'
        statement[field]='now'
        return statement
    


class Mail:
    def __init__(self, from_text=None, destination=None, subject=None, custom_data=None, link=None):
        self.from_text = from_text;
        self.destination = destination;
        self.subject = subject;
        self.custom_data = custom_data;
        self.link = link;
    
    def sendgrid_client(self, _api_key=None):
        api_key = _api_key;
        if api_key is None:
            api_key = SENDGRID_API_KEY;
        return SendGridAPIClient(api_key);

    def get_user(self):
        user_model = User().query([{"key": "email", "value": self.destination, "comp": "=="}], True);
        if user_model is not None and user_model.uid is None:
            return None;
        return user_model;

    def check_time_send(self, check=True, field=VALIDATION_FIELD):
        user_model = self.get_user()
        if user_model is not None and user_model.uid is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error('User not exist', 'USER_NOT_EXIST'))
        if check:
            print('field', field)
            time_sended = getTimeSended(user_model, field=field);
            if time_sended['status']=='error':
                return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error('User not exist', 'USER_NOT_EXIST'))
            if time_sended[field]!='now':
                compare = compareTimes(time_sended[field]);
                if compare['status'] == 'error':
                    minutes = DEFAULT_EMAIL_TIME_TO_RESEND - compare['difference']
                    return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, {"minutes": minutes}, Error(f"Please wait {minutes} minutes and retry", 'WAIT_FEW_MINUTES'))

        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, user_model, None)
    
    def sendMailRegistration(self, domain, _api_key=None, _template_id=None, bccs=[]):
        if domain is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error('Invalid domain', 'INVALID_REQUEST'))
        result = None
        template_id = _template_id;
        if template_id is None:
            template_id = REGISTRATION_TEMPLATE_ID
        message = Mail(
            from_email=self.from_text+'@'+domain,
            to_emails=[self.destination],
            )
        try:
            message.template_id = template_id
            personalization = Personalization()
            for bcc in bccs:
                personalization.add_bcc(Email(bcc))
            message.add_personalization(personalization)
            response = self.sendgrid_client(_api_key=_api_key).send(message)
            if response.status_code==200 or response.status_code==202:
                result = ApiResponse(ResponseStatus.OK, response.status_code, {"message": "Email sent successfully"}, None)
                print(result)
            else:
                result = ApiResponse(ResponseStatus.ERROR, response.status_code, None, Error('An error occurated', 'INVALID_REQUEST'))
        except Exception as e:
            print(e)
            result = ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(str(e), 'INVALID_REQUEST'))
        
        return result
    
    def sendMailPaymentSuccess(self, domain, _api_key=None, _template_id=None, bccs=[]):
        if domain is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error('Invalid domain', 'INVALID_REQUEST'))
        result = None
        message = SendGridMail()
        try:
            template_id = _template_id;
            if template_id is None:
                template_id = PAYMENT_TEMPLATE_ID
            message.from_email = self.from_text+'@'+domain
            message.template_id = template_id
            personalization = Personalization()
            personalization.dynamic_template_data = self.custom_data
            personalization.add_to(Email(self.destination))
            personalization.subject = self.subject
            for bcc in bccs:
                personalization.add_bcc(Email(bcc))
            message.add_personalization(personalization)
        
            response = self.sendgrid_client(_api_key).send(message) 
            if response.status_code==200 or response.status_code==202:
                result = ApiResponse(ResponseStatus.OK, response.status_code, {"message": "Email sent successfully"}, None)
            else:
                result = ApiResponse(ResponseStatus.ERROR, response.status_code, None, Error('An error occurated', 'INVALID_REQUEST'))
        except Exception as e:
            print(e)
            result = ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(str(e), 'INVALID_REQUEST'))
        
        return result
    
    def sendMailEmailVerfication(self, domain, _api_key=None, _template_id=None, bccs=[], check=False, field=VALIDATION_FIELD) -> 'ApiResponse':
        if domain is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error('Invalid domain', 'INVALID_REQUEST'))
        result = None
        message = SendGridMail(
            from_email=self.from_text+'@'+domain,
            subject=self.subject
            )
        time_sended = self.check_time_send(check);
        if time_sended.status==ResponseStatus.ERROR:
            return time_sended
        u: User = time_sended.data
        user_model = User(u.to_json())
        try:
            template_id = _template_id;
            if template_id is None:
                template_id = os.environ.get('EMAIL_VERIFICATION_TEMPLATE_ID')
            custom_data = self.custom_data;
            if self.link is not None:
                c_data = {'link': urllib.parse.unquote(self.link)}
                if custom_data is not None:
                    custom_data = {**custom_data, **c_data, "subject": self.subject}
                else:
                    custom_data = {**c_data, "subject": self.subject}
            print('message subject ===>', self.subject)
            message.subject = self.subject
            message.from_email = self.from_text+'@'+domain
            message.template_id = template_id
            personalization = Personalization()
            if self.custom_data is not None:
                personalization.dynamic_template_data = custom_data
            personalization.add_to(Email(email=self.destination, subject=self.subject))
            personalization.subject = self.subject
            for bcc in bccs:
                personalization.add_bcc(Email(bcc))
            message.add_personalization(personalization)
            email_test_mode = boolean(os.environ.get('EMAIL_TEST_MODE'))
            if email_test_mode is None or email_test_mode is True:
                response = Object(status_code=200, message="success")
            else:
                response = self.sendgrid_client(_api_key).send(message) 
            if response.status_code==200 or response.status_code==202:
                setTimeSended(user_model, field=field);
                result = ApiResponse(ResponseStatus.OK, response.status_code, {"message": "Email sent successfully"}, None)
            else:
                result = ApiResponse(ResponseStatus.ERROR, response.status_code, None, Error('An error occurated that\'s why your email were not sent', 'EMAIL_NOT_SENT', 1))
        except Exception as e:
            print(e)
            result = ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(str(e), 'INVALID_REQUEST'))
        
        return result
    

class MailSetting(Mail):
    DEFAULT_CODE_VALIDATION_BODY = "Vous êtes prêts pour commencer. Saisissez le code ci dessous pour valider votre email."
    DEFAULT_RESET_PSW_BODY = "Cliquez sur le lien ci dessous pour réinitialiser votre mot de passe "
    CODE_VALIDATION_TEMPLATE_ID = os.environ.get('CODE_VALIDATION_TEMPLATE_ID')
    RESET_PASSWORD_TEMPLATE_ID = os.environ.get('RESET_PASSWORD_TEMPLATE_ID')
    def __init__(self, from_text=None, destination=None, subject=None, custom_data=None, link=None):
        super().__init__(from_text, destination, subject, custom_data, link);
    
    def generate_code(self, code_type="email_validation", field=VALIDATION_FIELD):
        user_model = self.get_user();
        code = None;
        if user_model is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error('User not found', 'INVALID_REQUEST'))
        init = Code({"id": Code.get_id(None, target=user_model.uid, type=code_type)}).getCode();
        time_sended = None;
        u: User = None;
        if init is not None:
            time_sended = self.check_time_send(field=field);
            if time_sended.status==ResponseStatus.ERROR:
                return time_sended
            else:
                if init.is_expired():
                    u = time_sended.data;
                else:
                    code = init;
        else:
            u = user_model
        if code is None:
            user = User(u.to_json())
            uid = user.uid;
            data_code = str(generate_random_code());
            model_ = {
                "target": uid,
                "code_type": code_type,
                "code": data_code
            }
            id = getattr(Code.generate(**model_).data, 'id',None);
            code_model = Code();
            code_model.id = id;
            code_model.code = data_code;
            code_model.code_type = code_type
            code_model.expires_in = 600
            code_model.target = uid;
            code = code_model.getCode()
            if code is None or code.is_expired():
                if code_type=='password_reset':
                    code_model.code = Crypto().hash(code_model.code);
                create = Code().create(**code_model.to_json());
                return create
            
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, code.to_json(), None)

    def generate_email_validation_code(self, code_type="email_validation", domain=None, bccs=[], template_id=None, _body=None, url=None, query_params=None):
        field = VALIDATION_FIELD;
        link_url = None;
        if code_type == 'password_reset':
            field = PSW_RESET_FIELD
            if template_id is None:
                template_id = self.RESET_PASSWORD_TEMPLATE_ID;
        else:
            if template_id is None:
                template_id = self.CODE_VALIDATION_TEMPLATE_ID;
        code_request = self.generate_code(code_type=code_type, field=field) 
        print('code request', code_request)
        if code_request.status==ResponseStatus.ERROR:
            return link_url, code_request;
        code = Code(code_request.data)
        # print('code data', code_request.data)
        body = _body;
        if body is None:
            if field==PSW_RESET_FIELD:
                body = self.DEFAULT_RESET_PSW_BODY
            else:
                body = self.DEFAULT_CODE_VALIDATION_BODY
        self.custom_data = {"body": body}
        if field == VALIDATION_FIELD:
            self.custom_data['code']=code.code
            link_url = code.code
        elif field == PSW_RESET_FIELD:
            if url is None:
                url = os.environ.get('BASE_URL')
            query = {"code": code.code, "target": code.target};
            if query_params is  not None:
                query = {**query, **query_params}
            params = urllib.parse.urlencode(query=query)
            link_url = f"{url}/reset/?{params}";
            self.custom_data['link'] = link_url
            print('link', link_url)
        print('code', code.code)
        response = ApiResponse(ResponseStatus.OK, 200, {"message": "Email sent successfully", "code": code.code}, None)
        email_test_mode = boolean(os.environ.get('EMAIL_TEST_MODE'))
        if bool(email_test_mode) is False:
            # print('sending email', self.subject)
            response = self.sendMailEmailVerfication(domain=domain, _template_id=template_id, bccs=bccs, check=False, field=field)
        return link_url, response


from dataclasses import dataclass
from  ....utils.utils import BaseFieldsClass
from google.cloud.firestore_v1.document import Timestamp
from google.api_core.datetime_helpers import DatetimeWithNanoseconds
@dataclass
class ClientCustomerIdFields(BaseFieldsClass):
    accountId = "accountId"
    customerId = "customerId"
    name = "name"
    enabled = "enabled"
    owner = "owner"
    creationDate = "creationDate"
    creationDateString = "creationDateString"
    createdAt = "createdAt"
    isDefault = "isDefault"
    def __init__(self, **kwargs):
        self.class_fields_props = ClientCustomerIdFieldsProps
        super().__init__(**kwargs)

ClientCustomerIdFieldsProps = {
    ClientCustomerIdFields.accountId: {
        "internal": True,
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ClientCustomerIdFields.name: {
        "internal": True,
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ClientCustomerIdFields.customerId: {
        "internal": True,
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ClientCustomerIdFields.owner: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": False,
        "default_value": None
    },
    ClientCustomerIdFields.enabled: {
        "internal": True,
        "type": bool,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": True
    },
    ClientCustomerIdFields.isDefault: {
        "internal": True,
        "type": bool,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": False
    },
    ClientCustomerIdFields.createdAt: {
        "internal": False,
        "type": [int, str, Timestamp, DatetimeWithNanoseconds],
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ClientCustomerIdFields.creationDate: {
        "internal": True,
        "type": int,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ClientCustomerIdFields.creationDateString: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
}


CLIENT_CUSTOMER_ID_PICKABLE_FIELDS = ClientCustomerIdFields(fields_props=ClientCustomerIdFieldsProps).filtered_keys('pickable', True)



from .ad import (ResponsiveDisplayAd, filter_responsive_display_ad_request_fields, NativeAdsToPublish)
from .ad_fields import (ResponsiveDisplayAdFields, ResponsiveDisplayAdFieldsProps, RESPONSIVE_DISPLAY_AD_PICKABLE_FIELDS, RESPONSIVE_DISPLAY_ADS_COLLECTION)



from dataclasses import dataclass
from .....utils import DictUtils
import os

MFA_COLLECTION = os.environ.get('MFA_COLLECTION', 'mfa_credentials')


@dataclass
class MFAFields:
    id = "id"
    user_uid = "user_uid"
    method = "method"
    totp_secret = "totp_secret"
    phone = "phone"
    backup_codes = "backup_codes"
    is_enabled = "is_enabled"
    enrolled_at = "enrolled_at"
    last_used_at = "last_used_at"

    @staticmethod
    def keys():
        return DictUtils.get_keys(MFAFieldsProps)

    @staticmethod
    def filtered_keys(field, condition=True):
        mutable = DictUtils.filter(MFAFieldsProps, DictUtils.get_keys(MFAFieldsProps), field, condition)
        return DictUtils.get_keys(mutable)


MFAFieldsProps = {
    MFAFields.id: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    MFAFields.user_uid: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    MFAFields.method: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "totp",
        "pickable": True
    },
    MFAFields.totp_secret: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    MFAFields.phone: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    MFAFields.backup_codes: {
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": [],
        "pickable": True
    },
    MFAFields.is_enabled: {
        "type": bool,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": False,
        "pickable": True
    },
    MFAFields.enrolled_at: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    MFAFields.last_used_at: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
}

STANDARD_FIELDS = MFAFields.filtered_keys('pickable', True)



objectives_data = [
    {
        "obj_id": '1',
        "primary": {
    "icon": 'local_offer',
    "title": 'sales',
    },
    "secondary": {
    "icon": 'local_offer',
    "title": 'sales',
    "description": 'sales_description',
    },
},
{
    "obj_id": '2',
    "primary": {
    "icon": 'people',
    "title": 'prospects',
    },
    "secondary": {
    "icon": 'people',
    "title": 'prospects',
    "description": 'prospect_description',
    },
},
{
    "obj_id": '3',
    "primary": {
    "icon": 'web',
    "title": 'traffic',
    },
    "secondary": {
    "icon": 'web',
    "title": 'traffic',
    "description": 'traffic_description',
    },
},
{
    "obj_id": '4',
    "primary": {
    "icon": 'auto_awesome',
    "title": 'brand_interest',
    },
    "secondary": {
    "icon": 'auto_awesome',
    "title": 'brand_interest',
    "description": 'brand_interest_description',
    },
},
{
    "obj_id": '5',
    "primary": {
    "icon": 'stars',
    "title": 'audience_brand_awareness',
    },
    "secondary": {
    "icon": 'stars',
    "title": 'audience_brand_awareness',
    "description": 'audience_brand_awareness_description',
    },
},
{
    "obj_id": '6',
    "primary": {
    "icon": 'settings',
    "title": 'no_objective',
    },
    "secondary": {
    "icon": 'settings',
    "title": 'no_objective',
    "description": 'unspecified_objective',
    },
}
]


import hashlib
import json
import pickle
import threading
import time
from collections import OrderedDict
from datetime import datetime, timedelta
from typing import Any, Callable, Dict, List, Optional, Union


class CacheItem:
    """Represents a single cache item with metadata."""
    
    def __init__(self, key: str, value: Any, ttl: Optional[int] = None):
        self.key = key
        self.value = value
        self.created_at = time.time()
        self.last_accessed = time.time()
        self.access_count = 0
        self.ttl = ttl  # Time to live in seconds
    
    def is_expired(self) -> bool:
        """Check if the cache item has expired."""
        if self.ttl is None:
            return False
        return time.time() - self.created_at > self.ttl
    
    def access(self):
        """Mark the item as accessed."""
        self.last_accessed = time.time()
        self.access_count += 1
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert cache item to dictionary for serialization."""
        return {
            'key': self.key,
            'value': self.value,
            'created_at': self.created_at,
            'last_accessed': self.last_accessed,
            'access_count': self.access_count,
            'ttl': self.ttl
        }


class Cache:
    """
    A comprehensive in-memory caching system with multiple eviction policies,
    TTL support, statistics, and persistence capabilities.
    """
    
    def __init__(self, 
                 max_size: int = 1000,
                 eviction_policy: str = 'lru',
                 default_ttl: Optional[int] = None,
                 enable_stats: bool = True,
                 enable_persistence: bool = False,
                 persistence_file: str = 'cache_data.pkl'):
        """
        Initialize the cache.
        
        Args:
            max_size: Maximum number of items in cache
            eviction_policy: 'lru', 'lfu', 'fifo', or 'random'
            default_ttl: Default time to live in seconds
            enable_stats: Enable cache statistics
            enable_persistence: Enable cache persistence
            persistence_file: File path for persistence
        """
        self.max_size = max_size
        self.eviction_policy = eviction_policy
        self.default_ttl = default_ttl
        self.enable_stats = enable_stats
        self.enable_persistence = enable_persistence
        self.persistence_file = persistence_file
        
        # Core cache storage
        self._cache: Dict[str, CacheItem] = {}
        self._access_order = OrderedDict()  # For LRU/FIFO
        self._access_frequency: Dict[str, int] = {}  # For LFU
        
        # Threading support
        self._lock = threading.RLock()
        
        # Statistics
        self._stats = {
            'hits': 0,
            'misses': 0,
            'sets': 0,
            'deletes': 0,
            'evictions': 0,
            'expirations': 0
        }
        
        # Background cleanup thread
        self._cleanup_thread = None
        self._stop_cleanup = False
        self._start_cleanup_thread()
        
        # Load from persistence if enabled
        if self.enable_persistence:
            self._load_from_persistence()
    
    def set(self, key: str, value: Any, ttl: Optional[int] = None) -> bool:
        """
        Set a value in the cache.
        
        Args:
            key: Cache key
            value: Value to cache
            ttl: Time to live in seconds (overrides default_ttl)
        
        Returns:
            True if successful, False otherwise
        """
        with self._lock:
            try:
                # Use provided TTL or default
                item_ttl = ttl if ttl is not None else self.default_ttl
                
                # Create cache item
                cache_item = CacheItem(key, value, item_ttl)
                
                # Check if we need to evict items
                if len(self._cache) >= self.max_size:
                    self._evict_item()
                
                # Store the item
                self._cache[key] = cache_item
                self._access_order[key] = time.time()
                self._access_frequency[key] = 0
                
                # Update statistics
                if self.enable_stats:
                    self._stats['sets'] += 1
                
                # Persist if enabled
                if self.enable_persistence:
                    self._save_to_persistence()
                
                return True
                
            except Exception as e:
                print(f"Error setting cache item: {e}")
                return False
    
    def get(self, key: str, default: Any = None) -> Any:
        """
        Get a value from the cache.
        
        Args:
            key: Cache key
            default: Default value if key not found
        
        Returns:
            Cached value or default
        """
        with self._lock:
            try:
                if key in self._cache:
                    item = self._cache[key]
                    
                    # Check if expired
                    if item.is_expired():
                        self._remove_item(key)
                        if self.enable_stats:
                            self._stats['expirations'] += 1
                        return default
                    
                    # Update access information
                    item.access()
                    self._access_order[key] = time.time()
                    self._access_frequency[key] = self._access_frequency.get(key, 0) + 1
                    
                    # Update statistics
                    if self.enable_stats:
                        self._stats['hits'] += 1
                    
                    return item.value
                else:
                    # Update statistics
                    if self.enable_stats:
                        self._stats['misses'] += 1
                    return default
                    
            except Exception as e:
                print(f"Error getting cache item: {e}")
                return default
    def getCache(self, key: str, cls = None):
        if bool(key) is False:
            return None
        value = self.get(key)
        if value is not None and cls is not None:
            value = cls.from_dict(value)
            return value
        return value
    
    def delete(self, key: str) -> bool:
        """
        Delete a key from the cache.
        
        Args:
            key: Cache key to delete
        
        Returns:
            True if deleted, False if not found
        """
        with self._lock:
            try:
                if key in self._cache:
                    self._remove_item(key)
                    
                    # Update statistics
                    if self.enable_stats:
                        self._stats['deletes'] += 1
                    
                    # Persist if enabled
                    if self.enable_persistence:
                        self._save_to_persistence()
                    
                    return True
                return False
                
            except Exception as e:
                print(f"Error deleting cache item: {e}")
                return False
    
    def exists(self, key: str) -> bool:
        """Check if a key exists in the cache."""
        with self._lock:
            if key in self._cache:
                item = self._cache[key]
                if item.is_expired():
                    self._remove_item(key)
                    return False
                return True
            return False
    
    def clear(self) -> bool:
        """Clear all items from the cache."""
        with self._lock:
            try:
                self._cache.clear()
                self._access_order.clear()
                self._access_frequency.clear()
                
                # Update statistics
                if self.enable_stats:
                    self._stats['deletes'] += len(self._cache)
                
                # Persist if enabled
                if self.enable_persistence:
                    self._save_to_persistence()
                
                return True
                
            except Exception as e:
                print(f"Error clearing cache: {e}")
                return False
    
    def keys(self) -> List[str]:
        """Get all valid (non-expired) keys in the cache."""
        with self._lock:
            valid_keys = []
            expired_keys = []
            
            for key, item in self._cache.items():
                if item.is_expired():
                    expired_keys.append(key)
                else:
                    valid_keys.append(key)
            
            # Remove expired keys
            for key in expired_keys:
                self._remove_item(key)
            
            return valid_keys
    
    def size(self) -> int:
        """Get the current number of items in the cache."""
        with self._lock:
            # Clean expired items first
            self._cleanup_expired()
            return len(self._cache)
    
    def get_stats(self) -> Dict[str, Any]:
        """Get cache statistics."""
        with self._lock:
            stats = self._stats.copy()
            stats['current_size'] = len(self._cache)
            stats['max_size'] = self.max_size
            stats['hit_rate'] = (stats['hits'] / (stats['hits'] + stats['misses'])) if (stats['hits'] + stats['misses']) > 0 else 0
            return stats
    
    def reset_stats(self):
        """Reset cache statistics."""
        with self._lock:
            self._stats = {
                'hits': 0,
                'misses': 0,
                'sets': 0,
                'deletes': 0,
                'evictions': 0,
                'expirations': 0
            }
    
    def set_ttl(self, key: str, ttl: int) -> bool:
        """
        Set TTL for an existing key.
        
        Args:
            key: Cache key
            ttl: New TTL in seconds
        
        Returns:
            True if successful, False if key not found
        """
        with self._lock:
            if key in self._cache:
                self._cache[key].ttl = ttl
                self._cache[key].created_at = time.time()
                return True
            return False
    
    def get_ttl(self, key: str) -> Optional[int]:
        """Get remaining TTL for a key in seconds."""
        with self._lock:
            if key in self._cache:
                item = self._cache[key]
                if item.ttl is None:
                    return None
                remaining = item.ttl - (time.time() - item.created_at)
                return max(0, int(remaining))
            return None
    
    def touch(self, key: str) -> bool:
        """Update the last access time for a key."""
        with self._lock:
            if key in self._cache:
                item = self._cache[key]
                if not item.is_expired():
                    item.access()
                    self._access_order[key] = time.time()
                    return True
            return False
    
    def increment(self, key: str, delta: int = 1, default: int = 0) -> Optional[int]:
        """
        Increment a numeric value in the cache.
        
        Args:
            key: Cache key
            delta: Amount to increment by
            default: Default value if key doesn't exist
        
        Returns:
            New value or None if not numeric
        """
        with self._lock:
            try:
                current = self.get(key, default)
                if isinstance(current, (int, float)):
                    new_value = current + delta
                    self.set(key, new_value)
                    return new_value
                return None
            except Exception:
                return None
    
    def decrement(self, key: str, delta: int = 1, default: int = 0) -> Optional[int]:
        """Decrement a numeric value in the cache."""
        return self.increment(key, -delta, default)
    
    def get_many(self, keys: List[str]) -> Dict[str, Any]:
        """Get multiple values from the cache."""
        result = {}
        for key in keys:
            value = self.get(key)
            if value is not None:
                result[key] = value
        return result
    
    def set_many(self, data: Dict[str, Any], ttl: Optional[int] = None) -> bool:
        """Set multiple values in the cache."""
        try:
            for key, value in data.items():
                self.set(key, value, ttl)
            return True
        except Exception:
            return False
    
    def delete_many(self, keys: List[str]) -> int:
        """Delete multiple keys from the cache."""
        deleted_count = 0
        for key in keys:
            if self.delete(key):
                deleted_count += 1
        return deleted_count
    
    def _evict_item(self):
        """Evict an item based on the eviction policy."""
        if not self._cache:
            return
        
        key_to_evict = None
        
        if self.eviction_policy == 'lru':
            # Least Recently Used
            key_to_evict = next(iter(self._access_order))
        elif self.eviction_policy == 'lfu':
            # Least Frequently Used
            key_to_evict = min(self._access_frequency.items(), key=lambda x: x[1])[0]
        elif self.eviction_policy == 'fifo':
            # First In First Out
            key_to_evict = next(iter(self._access_order))
        elif self.eviction_policy == 'random':
            # Random selection
            import random
            key_to_evict = random.choice(list(self._cache.keys()))
        
        if key_to_evict:
            self._remove_item(key_to_evict)
            if self.enable_stats:
                self._stats['evictions'] += 1
    
    def _remove_item(self, key: str):
        """Remove an item from the cache."""
        if key in self._cache:
            del self._cache[key]
        if key in self._access_order:
            del self._access_order[key]
        if key in self._access_frequency:
            del self._access_frequency[key]
    
    def _cleanup_expired(self):
        """Remove all expired items from the cache."""
        expired_keys = []
        for key, item in self._cache.items():
            if item.is_expired():
                expired_keys.append(key)
        
        for key in expired_keys:
            self._remove_item(key)
            if self.enable_stats:
                self._stats['expirations'] += 1
    
    def _start_cleanup_thread(self):
        """Start background cleanup thread."""
        def cleanup_worker():
            while not self._stop_cleanup:
                try:
                    with self._lock:
                        self._cleanup_expired()
                    time.sleep(60)  # Clean up every minute
                except Exception:
                    pass
        
        self._cleanup_thread = threading.Thread(target=cleanup_worker, daemon=True)
        self._cleanup_thread.start()
    
    def _save_to_persistence(self):
        """Save cache data to persistence file."""
        if not self.enable_persistence:
            return
        
        try:
            # Convert cache items to serializable format
            serializable_cache = {}
            for key, item in self._cache.items():
                if not item.is_expired():
                    serializable_cache[key] = item.to_dict()
            
            data = {
                'cache': serializable_cache,
                'access_order': dict(self._access_order),
                'access_frequency': self._access_frequency,
                'stats': self._stats
            }
            
            with open(self.persistence_file, 'wb') as f:
                pickle.dump(data, f)
                
        except Exception as e:
            print(f"Error saving cache to persistence: {e}")
    
    def _load_from_persistence(self):
        """Load cache data from persistence file."""
        if not self.enable_persistence:
            return
        
        try:
            with open(self.persistence_file, 'rb') as f:
                data = pickle.load(f)
            
            # Restore cache items
            for key, item_data in data.get('cache', {}).items():
                item = CacheItem(
                    key=item_data['key'],
                    value=item_data['value'],
                    ttl=item_data['ttl']
                )
                item.created_at = item_data['created_at']
                item.last_accessed = item_data['last_accessed']
                item.access_count = item_data['access_count']
                
                self._cache[key] = item
            
            # Restore other data
            self._access_order = OrderedDict(data.get('access_order', {}))
            self._access_frequency = data.get('access_frequency', {})
            self._stats = data.get('stats', self._stats)
            
        except FileNotFoundError:
            # No persistence file exists yet
            pass
        except Exception as e:
            print(f"Error loading cache from persistence: {e}")
    
    def __enter__(self):
        """Context manager entry."""
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit."""
        self.close()
    
    def close(self):
        """Clean up resources."""
        self._stop_cleanup = True
        if self._cleanup_thread and self._cleanup_thread.is_alive():
            self._cleanup_thread.join(timeout=1)
        
        if self.enable_persistence:
            self._save_to_persistence()
    
    def __len__(self) -> int:
        """Return the number of items in the cache."""
        return self.size()
    
    def __contains__(self, key: str) -> bool:
        """Check if a key exists in the cache."""
        return self.exists(key)
    
    def __getitem__(self, key: str) -> Any:
        """Get item using bracket notation."""
        value = self.get(key)
        if value is None:
            raise KeyError(key)
        return value
    
    def __setitem__(self, key: str, value: Any):
        """Set item using bracket notation."""
        self.set(key, value)
    
    def __delitem__(self, key: str):
        """Delete item using bracket notation."""
        if not self.delete(key):
            raise KeyError(key)


# Convenience functions for common use cases
def create_cache(max_size: int = 1000, 
                eviction_policy: str = 'lru',
                default_ttl: Optional[int] = None) -> Cache:
    """Create a new cache instance with common settings."""
    return Cache(max_size=max_size, 
                eviction_policy=eviction_policy,
                default_ttl=default_ttl)


def create_persistent_cache(persistence_file: str = 'cache_data.pkl',
                          max_size: int = 1000,
                          eviction_policy: str = 'lru',
                          default_ttl: Optional[int] = None) -> Cache:
    """Create a new persistent cache instance."""
    return Cache(max_size=max_size,
                eviction_policy=eviction_policy,
                default_ttl=default_ttl,
                enable_persistence=True,
                persistence_file=persistence_file)


import uuid
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Optional

from adafri.v1.base.firebase_collection import FirebaseCollectionBase

DOCUMENT_COLLECTION = 'workflow_documents'

DOCUMENT_PICKABLE_FIELDS = [
    'id', 'workspaceId', 'title', 'icon', 'content',
    'isPublished', 'allowedTeamIds', 'allowedMemberUids',
    'createdAt', 'updatedAt', 'createdBy', 'lastEditedBy',
]

DOCUMENT_MUTABLE_FIELDS = [
    'title', 'icon', 'content', 'isPublished', 'lastEditedBy',
    'allowedTeamIds', 'allowedMemberUids',
]


def _now() -> str:
    return datetime.now(timezone.utc).isoformat()


def _new_id() -> str:
    return str(uuid.uuid4())


class WorkflowDocumentFields:
    @staticmethod
    def check_create_fields(data: dict) -> dict:
        if not data:
            return {'error': 'Missing data'}
        if not data.get('workspaceId', '').strip():
            return {'error': 'workspaceId is required'}
        return data

    @staticmethod
    def mutable_keys() -> list:
        return DOCUMENT_MUTABLE_FIELDS


@dataclass(init=False)
class WorkflowDocument(FirebaseCollectionBase):
    id: str
    workspaceId: str
    title: str
    icon: str
    content: str
    isPublished: bool
    allowedTeamIds: list
    allowedMemberUids: list
    createdAt: str
    updatedAt: str
    createdBy: str
    lastEditedBy: str

    def __init__(self, data=None, **kwargs):
        super().__init__(collection_name=DOCUMENT_COLLECTION, **kwargs)
        d = data or {}
        self.id = d.get('id') or _new_id()
        self.workspaceId = d.get('workspaceId', '')
        self.title = d.get('title', 'Untitled')
        self.icon = d.get('icon', '📄')
        self.content = d.get('content', '')
        self.isPublished = bool(d.get('isPublished', False))
        self.allowedTeamIds = list(d.get('allowedTeamIds') or [])
        self.allowedMemberUids = list(d.get('allowedMemberUids') or [])
        self.createdAt = d.get('createdAt') or _now()
        self.updatedAt = d.get('updatedAt') or _now()
        self.createdBy = d.get('createdBy', '')
        self.lastEditedBy = d.get('lastEditedBy', '')

    @classmethod
    def from_dict(cls, obj: Any, **kwargs) -> 'WorkflowDocument':
        return cls(obj, **kwargs)

    def to_dict(self) -> dict:
        return {
            'id': self.id,
            'workspaceId': self.workspaceId,
            'title': self.title,
            'icon': self.icon,
            'content': self.content,
            'isPublished': self.isPublished,
            'allowedTeamIds': self.allowedTeamIds,
            'allowedMemberUids': self.allowedMemberUids,
            'createdAt': self.createdAt,
            'updatedAt': self.updatedAt,
            'createdBy': self.createdBy,
            'lastEditedBy': self.lastEditedBy,
        }

    @staticmethod
    def toListDict(items: list, fields=None) -> list:
        return [item.to_dict() for item in items]

    @classmethod
    def find(cls, doc_id: str) -> 'Optional[WorkflowDocument]':
        inst = cls()
        doc = inst.collection().document(doc_id).get()
        if not doc.exists:
            return None
        return cls({**doc.to_dict(), 'id': doc.id})

    @classmethod
    def listByWorkspace(cls, workspace_id: str) -> 'list[WorkflowDocument]':
        inst = cls()
        docs = (
            inst.collection()
            .where('workspaceId', '==', workspace_id)
            .order_by('updatedAt', direction='DESCENDING')
            .stream()
        )
        return [cls({**d.to_dict(), 'id': d.id}) for d in docs]

    def save(self) -> 'WorkflowDocument':
        self.collection().document(self.id).set(self.to_dict())
        return self

    def update(self, data: dict) -> 'WorkflowDocument':
        filtered = {k: v for k, v in data.items() if k in DOCUMENT_MUTABLE_FIELDS}
        filtered['updatedAt'] = _now()
        self.collection().document(self.id).update(filtered)
        for k, v in filtered.items():
            setattr(self, k, v)
        return self

    def delete(self) -> None:
        self.collection().document(self.id).delete()



from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any

from .....utils import DictUtils, Crypto, get_object_model_class, init_class_kwargs
from ....base.firebase_collection import FirebaseCollectionBase, getTimestamp
from .audit_log_fields import AuditLogFields, AuditLogFieldsProps, STANDARD_FIELDS, AUDIT_LOG_COLLECTION
from .....utils.response import ApiResponse, Error, ResponseStatus, StatusCode


@dataclass(init=False)
class AuditLog(FirebaseCollectionBase):
    id: str
    event_type: str
    user_uid: str
    client_id: str
    ip_address: str
    user_agent: str
    metadata: dict
    created_at: str

    def __init__(self, log=None, **kwargs):
        if type(log) is str:
            log = {"event_type": log}
        (cls_object, keys, data_args) = init_class_kwargs(self, log, STANDARD_FIELDS, AuditLogFieldsProps, AUDIT_LOG_COLLECTION, ['id'], **kwargs)
        super().__init__(**data_args)
        for key in keys:
            setattr(self, key, cls_object[key])

    @staticmethod
    def generate_model(_key_="default_value"):
        model = {}
        props = AuditLogFieldsProps
        for k in DictUtils.get_keys(props):
            model[k] = props[k][_key_]
        return model

    @staticmethod
    def from_dict(log: Any = None, db=None, collection_name=None) -> 'AuditLog':
        cls_object, keys = get_object_model_class(log, AuditLog, AuditLogFieldsProps)
        return AuditLog(cls_object, db=db, collection_name=collection_name)

    def query(self, query_params: list, first=False, limit=None):
        try:
            result = []
            query_result = self.custom_query(query_params, first=first, limit=limit)
            if bool(query_result):
                if first:
                    return AuditLog.from_dict(log=query_result, db=self.db, collection_name=self.collection_name)
                else:
                    for doc in query_result:
                        result.append(AuditLog.from_dict(log=doc, db=self.db, collection_name=self.collection_name))
                    return result
            return None if first else []
        except Exception as e:
            print('Exception while querying AuditLog', e)
            return None if first else []

    def to_dict(self):
        return {
            "id": self.id,
            "event_type": self.event_type,
            "user_uid": self.user_uid,
            "client_id": self.client_id,
            "ip_address": self.ip_address,
            "user_agent": self.user_agent,
            "metadata": self.metadata,
            "created_at": self.created_at,
        }

    def to_json(self, _fields=None):
        return self.to_dict()

    def list_by_user(self, uid: str, limit: int = 50):
        return self.query([{"key": AuditLogFields.user_uid, "comp": "==", "value": uid}], first=False, limit=limit)

    @classmethod
    def log(cls, event_type: str, user_uid: str = "", client_id: str = "",
            ip_address: str = "", user_agent: str = "", metadata: dict = None):
        now = datetime.now(timezone.utc).isoformat()
        model = {
            "event_type": event_type,
            "user_uid": user_uid,
            "client_id": client_id,
            "ip_address": ip_address,
            "user_agent": user_agent,
            "metadata": metadata or {},
            "created_at": now,
        }
        obj = cls.from_dict(model)
        obj.id = Crypto().generate_id(event_type + "~" + user_uid + "~" + now)
        doc_ref = AuditLog(obj.to_json()).document_reference()
        doc_ref.set({**obj.to_json(), "createdAt": getTimestamp()}, merge=True)
        return obj



from ....base.firebase_collection import FirebaseCollectionBase
from .....utils import (DictUtils, init_class_kwargs, get_request_fields)
from .base_adgroup_fields import BaseAdGroupFieldsProps, STANDARD_FIELDS, BASE_ADGROUP_PICKABLE_FIELDS
from typing import Any
from dataclasses import dataclass
from adafri.utils import DateUtils
from ..ages import Ages
from ..genders import Genders
from ..devices import Devices
from google.cloud.firestore_v1.transforms import Sentinel
from datetime import datetime

def filter_request_fields(fields, default_fields = BASE_ADGROUP_PICKABLE_FIELDS):
    request_fields = get_request_fields(fields, default_fields, [default_fields[0]])
    if request_fields is None or bool(request_fields) is False:
        request_fields = [default_fields[0]]
    return request_fields



@dataclass(init=False)
class BaseAdGroup(FirebaseCollectionBase):
    id: str
    campaign_id: int
    name: str
    status: str
    ages: list[Ages]
    genders: list[Genders]
    devicesTargeted: list[Devices]
    devicesExcluded: list[Devices]
    is_removed: bool = False
    ad_group_id: int
    bid: float
    strategie: str
    owner: str
    provider: str
    clientCustomerId: str
    campaign_id_parent: str
    accountId: str
    createdAt: any
    createdBy: str
    api_version: str
    google_ads_ad_group: dict
    __baseFields =  BaseAdGroupFieldsProps

    def __init__(self, adgroup=None, collection_name=None, **kwargs):
        _adgroup = adgroup
        if type(adgroup) is str:
            _adgroup = {"id": adgroup}
        # if adgroup is not None:
        #     print('init first', adgroup['id'])
        (cls_object, keys, data_args) = init_class_kwargs(self, _adgroup, STANDARD_FIELDS, BaseAdGroupFieldsProps, collection_name, ['id'], **kwargs)
        super().__init__(**data_args);
        for key in keys:
            if key == "ages":
                setattr(self, key, Ages().fromListDict(cls_object[key], Ages))
            elif key == "genders":
                setattr(self, key, Genders().fromListDict(cls_object[key], Genders))
            elif key == "devicesTargeted":
                setattr(self, key, Devices().fromListDict(cls_object[key], Devices))
            elif key == "devicesExcluded":
                setattr(self, key, Devices().fromListDict(cls_object[key], Devices))
            elif key == "createdAt":
                createdAt = DateUtils.convert_firestore_timestamp_to_str(cls_object[key]);
                if isinstance(createdAt, Sentinel) is True:
                    createdAt = datetime.now().strftime('%Y-%m-%d %H:%M:%S');
                setattr(self, key, createdAt) 
            else:
                setattr(self, key, cls_object[key]) 


    @staticmethod
    def generate_model(_key_="default_value"):
        adgroup = {};
        props = BaseAdGroupFieldsProps
        for k in DictUtils.get_keys(props):
            adgroup[k] = props[k][_key_];
        return adgroup;
    def ages_to_list(self, ages: list[Ages] | None = None):
        if ages is None:
            return [];
        return Ages().toListDict(ages, None)
    def genders_to_list(self, genders: list[Genders] | None = None):
        if genders is None:
            return [];
        return Genders().toListDict(genders, None)
    def devices_to_list(self, devices: list[Devices] | None = None):
        if devices is None:
            return [];
        return Devices().toListDict(devices, None)
    
    def list_to_ages(self, ages: list[Ages] | None = None):
        if ages is None:
            return [];
        return Ages().fromListDict(ages, Ages)
    def list_to_genders(self, genders: list[Genders] | None = None):
        if genders is None:
            return [];
        return Genders().fromListDict(genders, Genders)
    def list_to_devices(self, devices: list[Devices] | None = None):
        if devices is None:
            return [];
        return Devices().fromListDict(devices, Devices)

    def is_published(self):
        return self.ad_group_id is not None and self.ad_group_id > 0
    def to_dict(self, fields=None):
        campaign_id = self.campaign_id
        ad_group_id = self.ad_group_id
        if campaign_id == 0:
            campaign_id = None;
        if ad_group_id == 0:
            ad_group_id = None;
        
        return DictUtils.remove_none_values({
            "id": self.id,
            "campaign_id": campaign_id,
            "name": self.name,
            "status": self.status,
            "ages": Ages().toListDict(self.ages, None),
            "genders": Genders().toListDict(self.genders, None),
            "devicesTargeted": Devices().toListDict(self.devicesTargeted, None),
            "devicesExcluded": Devices().toListDict(self.devicesExcluded, None),
            "is_removed": self.is_removed,
            "bid": self.bid,
            "strategie": self.strategie,
            "ad_group_id": ad_group_id,
            "owner": self.owner,
            "provider": self.provider,
            "clientCustomerId": self.clientCustomerId,
            "accountId": self.accountId,
            "campaign_id_parent": self.campaign_id_parent,
            "createdAt": DateUtils.convert_firestore_timestamp_to_str(self.createdAt),
            "createdBy": self.createdBy,
            "api_version": self.api_version,
            "google_ads_ad_group": self.google_ads_ad_group
        })
        


import os
from dataclasses import dataclass
from typing import Any

from google.cloud.firestore_v1.document import Timestamp

from .....utils.utils import (
    BaseClass,
    BaseFieldsClass,
    DictUtils,
    get_object_model_class,
    init_class_kwargs,
)
from ...models.adgroups.ads.base_ad_fields import BaseAdFields, BaseAdFieldsProps

DISPLAY_ADS_COLLECTION = os.environ.get('GADS_ADS_IMAGE_AD_COLLECTION');

@dataclass(init=False)
class ImageAdToPublishFields(BaseFieldsClass):
    id = "id"
    name = "name"
    mediaId = "mediaId"
    type = "type"
    width = "width"
    height = "height"
    urls = "urls"
    assetId = "assetId"
    url = "url"
    hash = "hash"
    previewUrl = "previewUrl"
    resourceName = "resourceName"
    sourceUrl = "sourceUrl"
    referenceId = "referenceId"
    type = "type"
    mimeType = "mimeType"
    owner = "owner"
    fileSize = "fileSize"
    clientCustomerId = "clientCustomerId"

@dataclass
class ImageAdBaseFields(BaseAdFields):
    mediaId = "mediaId"
    assetId = "assetId"
    url = "url"
    hash = "hash"
    mimeType = "mimeType"
    fileSize = "fileSize"
    previewUrl = "previewUrl"
    referenceId = "referenceId"
    type = "type"
    sourceUrl = "sourceUrl"
    urls = "urls"
    usedForRectangle = "usedForRectangle"
    usedForSquare = "usedForSquare"
    width = "width"
    height = "height"
    state = "state"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
@dataclass
class ImageAdFields(ImageAdBaseFields):
    ad_name = "ad_name"
    ad_type = "ad_type"
    url_image = "url_image"
    displayUrl = "displayUrl"
    finalAppUrls = "finalAppUrls"
    finalMobileUrls = "finalMobileUrls"
    finalUrls = "finalUrls"
    account = "account"
    adgroup = "adgroup"
    customer = "customer"
    google_ads_image_ad = "google_ads_image_ad"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.class_fields_props = ImageAdFieldsProps

BaseImageAdFieldsProps = {
    **BaseAdFieldsProps,
    ImageAdFields.mediaId: {
        "internal": True,
        "type": int,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ImageAdFields.assetId: {
        "internal": True,
        "type": int,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ImageAdFields.mimeType: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ImageAdFields.fileSize: {
        "internal": True,
        "type": [int, str],
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ImageAdFields.previewUrl: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ImageAdFields.referenceId: {
        "internal": True,
        "type": [str, int],
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ImageAdFields.type: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": "IMAGE"
    },
    ImageAdFields.sourceUrl: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ImageAdFields.state: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ImageAdFields.urls: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ImageAdFields.url: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ImageAdFields.hash: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ImageAdFields.usedForRectangle: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": False
    },
    ImageAdFields.usedForSquare: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": False
    },
    ImageAdFields.width: {
        "internal": True,
        "type": int,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ImageAdFields.height: {
        "internal": True,
        "type": int,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ImageAdFields.account: {
        "type": dict,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": False,
        "default_value": None
    },
    ImageAdFields.customer: {
        "type": dict,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": False,
        "default_value": None
    },
    ImageAdFields.adgroup: {
        "type": dict,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": False,
        "default_value": None
    },
}


ImageAdFieldsProps = {
    **BaseImageAdFieldsProps,
    ImageAdFields.ad_name: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ImageAdFields.ad_type: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ImageAdFields.url_image: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ImageAdFields.displayUrl: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ImageAdFields.finalAppUrls: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    ImageAdFields.finalMobileUrls: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    ImageAdFields.finalUrls: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    ImageAdFields.google_ads_image_ad: {
        "type": dict,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
}

STANDARD_FIELDS = ImageAdFields(fields_props=ImageAdFieldsProps).filtered_keys('pickable', True)
IMAGE_AD_PICKABLE_FIELDS = ImageAdFields(fields_props=ImageAdFieldsProps).filtered_keys('pickable', True)



from .country import Country

# from .file import (
#     crop_image,
#     crop_image_from_url,
#     estimate_pre_crop_file_size,
#     get_compression_estimate,
#     get_file_info_from_url,
#     get_file_size_from_url,
#     upload_image_from_file,
#     upload_image_from_url,
# )
from .phone_number import *
from .queries import *
from .request_helpers import *
from .response import *
from .utils import *



import uuid
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Optional

from adafri.v1.base.firebase_collection import FirebaseCollectionBase

CAMPAIGN_LINK_COLLECTION = 'workflow_campaign_links'

CAMPAIGN_LINK_PICKABLE_FIELDS = [
    'id', 'workspaceId', 'campaignId', 'linkedBy', 'linkedAt',
]


def _now() -> str:
    return datetime.now(timezone.utc).isoformat()


def _new_id() -> str:
    return str(uuid.uuid4())


@dataclass(init=False)
class WorkflowCampaignLink(FirebaseCollectionBase):
    id: str
    workspaceId: str
    campaignId: str
    linkedBy: str
    linkedAt: str

    def __init__(self, data=None, **kwargs):
        super().__init__(collection_name=CAMPAIGN_LINK_COLLECTION, **kwargs)
        d = data or {}
        self.id = d.get('id') or _new_id()
        self.workspaceId = d.get('workspaceId', '')
        self.campaignId = d.get('campaignId', '')
        self.linkedBy = d.get('linkedBy', '')
        self.linkedAt = d.get('linkedAt') or _now()

    @classmethod
    def from_dict(cls, obj: Any, **kwargs) -> 'WorkflowCampaignLink':
        return cls(obj, **kwargs)

    def to_dict(self) -> dict:
        return {
            'id': self.id,
            'workspaceId': self.workspaceId,
            'campaignId': self.campaignId,
            'linkedBy': self.linkedBy,
            'linkedAt': self.linkedAt,
        }

    @staticmethod
    def toListDict(items: list, fields=None) -> list:
        return [item.to_dict() for item in items]

    @classmethod
    def find(cls, link_id: str) -> 'Optional[WorkflowCampaignLink]':
        inst = cls()
        doc = inst.collection().document(link_id).get()
        if not doc.exists:
            return None
        return cls({**doc.to_dict(), 'id': doc.id})

    @classmethod
    def listByWorkspace(cls, workspace_id: str) -> 'list[WorkflowCampaignLink]':
        inst = cls()
        docs = (
            inst.collection()
            .where('workspaceId', '==', workspace_id)
            .order_by('linkedAt')
            .stream()
        )
        return [cls({**d.to_dict(), 'id': d.id}) for d in docs]

    def save(self) -> 'WorkflowCampaignLink':
        self.collection().document(self.id).set(self.to_dict())
        return self

    def delete(self) -> None:
        self.collection().document(self.id).delete()



from dataclasses import dataclass
from typing import Any
from  .....utils.utils import  BaseFieldsClass
from google.cloud.firestore_v1.document import Timestamp
from google.api_core.datetime_helpers import DatetimeWithNanoseconds

@dataclass
class BaseAdGroupFields(BaseFieldsClass):
    id = "id"
    name = "name"
    campaign_id = "campaign_id"
    ad_group_id = "ad_group_id"
    campaign_id_parent = "campaign_id_parent"
    status = "status"
    ages = "ages"
    genders = "genders"
    clientCustomerId = "clientCustomerId"
    servingStatus = "servingStatus"
    owner = "owner"
    accountId = "accountId"
    createdAt = "createdAt"
    createdBy = "createdBy"
    devicesTargeted = "devicesTargeted"
    devicesExcluded = "devicesExcluded"
    is_removed = "is_removed"
    provider = "provider"
    bid = "bid"
    strategie = "strategie"
    api_version = "api_version"
    google_ads_ad_group = "google_ads_ad_group"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



BaseAdGroupFieldsProps = {
    BaseAdGroupFields.id: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdGroupFields.name: {
        "internal": True,
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdGroupFields.status: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": "ENABLED"
    },
    BaseAdGroupFields.owner: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseAdGroupFields.provider: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": "https://adafri.app"
    },
    BaseAdGroupFields.accountId: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseAdGroupFields.campaign_id: {
        "internal": True,
        "type": int,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": 0
    },
    BaseAdGroupFields.campaign_id_parent: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseAdGroupFields.ad_group_id: {
        "internal": True,
        "type": int,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": 0
    },
    BaseAdGroupFields.ages: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    BaseAdGroupFields.genders: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    BaseAdGroupFields.clientCustomerId: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseAdGroupFields.createdBy: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": False,
        "default_value": None
    },
    BaseAdGroupFields.createdAt: {
        "type": [int, str, Timestamp, DatetimeWithNanoseconds],
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseAdGroupFields.is_removed: {
        "internal": True,
        "type": bool,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": False
    },
    BaseAdGroupFields.devicesTargeted: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    BaseAdGroupFields.devicesExcluded: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    BaseAdGroupFields.strategie: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": "CPM"
    },
    BaseAdGroupFields.bid: {
        "internal": True,
        "type": float,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": False,
        "default_value": 0.1
    },
    BaseAdGroupFields.api_version: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": "v0"
    },
    BaseAdGroupFields.google_ads_ad_group: {
        "internal": True,
        "type": dict,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    }
}


STANDARD_FIELDS = BaseAdGroupFields(fields_props=BaseAdGroupFieldsProps).filtered_keys('pickable', True)
BASE_ADGROUP_PICKABLE_FIELDS = BaseAdGroupFields(fields_props=BaseAdGroupFieldsProps).filtered_keys('pickable', True)



from adafri.cache.lib import cache
from adafri.utils import ApiResponse, ArrayUtils, Error, ResponseStatus, StatusCode
from adafri.v1.account import Account
from authlib.integrations.flask_client import OAuth
from authlib.integrations.flask_oauth2 import AuthorizationServer, ResourceProtector
from authlib.oauth2.rfc7636 import CodeChallenge
from authlib.oauth2.rfc6749 import grants as _grants

from ...oauth import JwtTokenGenerator, JwtTokenValidator, OAuthClient, OAuthToken
from ...oauth import OpenIDAuthorizationCodeGrant as AuthCodeGrant
from ...oauth import OpenIDCode, OpenIDHybridGrant
from ...oauth.models.token import TokenIntrospectionEndpoint, RefreshTokenGrant
from .jwks import jwks_manager

oidc_authorization_server = AuthorizationServer()
require_oidc_oauth = ResourceProtector()
import json
import os

import requests

flask_oauth = OAuth();
provider = None;

def get_adafri_provider_cfg(base=None):
    if base is not None:
        discovery_url = base + "/.well-known/openid-configuration"
        return requests.get(discovery_url).json()
    return None

def config_oidc_oauth(app, query_client=None, save_token=None, token_generators=[]):
    flask_oauth.init_app(app)
    if query_client is None:
        query_client = OAuthClient().get_by_client_id
    if save_token is None:
        save_token = OAuthToken().save
    oidc_authorization_server.query_client = query_client
    oidc_authorization_server.save_token = save_token
    oidc_authorization_server.init_app(app)
    oidc_authorization_server.register_grant(AuthCodeGrant, [
        OpenIDCode(require_nonce=False),
        CodeChallenge(required=True)
    ])
    # oidc_authorization_server.register_grant(OpenIDImplicitGrant)
    oidc_authorization_server.register_grant(OpenIDHybridGrant)
    oidc_authorization_server.register_grant(RefreshTokenGrant)
    oidc_authorization_server.register_grant(_grants.ClientCredentialsGrant)
    oidc_authorization_server.register_endpoint(TokenIntrospectionEndpoint)
    oidc_authorization_server.register_token_generator("default", JwtTokenGenerator(issuer=os.getenv('JWT_ISSUER')))
    oidc_authorization_server.register_token_generator("client_credentials", JwtTokenGenerator(issuer=os.getenv('JWT_ISSUER')))
    require_oidc_oauth.register_token_validator(JwtTokenValidator(issuer=os.getenv('JWT_ISSUER'),
        resource_server=os.getenv('JWT_ISSUER')))
        
def guards_oidc(scopes=['profile'], request=None):
    """
    A decorator that checks if the user has the required scopes to access this route.

    Args:
        scopes (list): A list of scopes that the user must have to access this route.
        Defaults to ['profile'].

    Returns:
        The user's token if the user has the required scopes, otherwise None.
    """
    with require_oidc_oauth.acquire(scopes=scopes) as token:
        return token
    
def get_admins():
    admins = []
    try:
        strings = os.getenv('ADMIN_USERS')
        if strings is None:
            return []
        admins = json.loads(strings)
    except:
        admins = []
    return admins

def is_admin(uid: str):
    admins = get_admins()
    if bool(admins) is True and uid in admins:
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, None, None);
    return ApiResponse(ResponseStatus.ERROR, StatusCode.status_401, None, Error("PERMISSION DENIED","INVALID_REQUEST", 1))
def guard_account(accountId: str, uid: str):
    admins = get_admins()
    account = cache.getCache(accountId, Account)
    if account is None:
        account = Account(account={"id": accountId}).get()
        if account is not None:
            cache.set(accountId, account.to_dict(), 225)
    if account is None:
        return ApiResponse(ResponseStatus.ERROR, StatusCode.status_401, None, Error("PERMISSION DENIED","INVALID_REQUEST", 1))
    if bool(admins) is True and uid in admins:
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, account, None);
    # account = ArrayUtils.find(accounts, lambda x: x.id == accountId)
    if account.owner != uid:
        links_cache = cache.get(f"links-{uid}")
        if links_cache is not None:
            links = Account().fromListDict(links_cache)
        else:
            links = Account(account=None).getUserAccountsLinked(uid)
            cache.set(f"links-{uid}", links, 225)
        # links = Account(account=None).getUserAccountsLinked(uid)
        # print('links', links)
        account = ArrayUtils.find(links, lambda x: x.id==accountId)
        if account is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_401, None, Error("PERMISSION DENIED","INVALID_REQUEST", 1))
    return ApiResponse(ResponseStatus.OK, StatusCode.status_200, account, None);


import os
from dataclasses import dataclass

from google.api_core.datetime_helpers import DatetimeWithNanoseconds
from google.cloud.firestore_v1.document import Timestamp

from ....utils.utils import BaseFieldsClass


@dataclass
class FileFields(BaseFieldsClass):
    id = "id"
    hash = 'hash'
    _id = '_id'
    accountId = 'accountId'
    blob_id = 'blob_id'
    blob_name = 'blob_name'
    bucket_name = 'bucket_name'
    colors = 'colors'
    content_type = 'content_type'
    created_time = 'created_time'
    directory_path = 'directory_path'
    extension = 'extension'
    file_source = 'file_source'
    file_url = 'file_url'
    url = "url"
    filename = 'filename'
    height = 'height'
    mime_type = 'mime_type'
    mode = 'mode'
    size = 'size'
    width = 'width'
    type = 'type'
    crop = "crop"
    crop_state = "crop_state"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.class_fields_props = FileFieldProps

FileFieldProps = {
    FileFields.id: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.hash: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields._id: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.accountId: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.blob_id: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.blob_name: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.bucket_name: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.colors: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.content_type: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.created_time: {
        "internal": True,
        "type": [int, str, Timestamp, DatetimeWithNanoseconds],
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.directory_path: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.extension: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.file_source: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.file_url: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.filename: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.height: {
        "internal": True,
        "type": int,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.mime_type: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.mode: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.size: {
        "internal": True,
        "type": int,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.type: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.width: {
        "internal": True,
        "type": int,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.url: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.crop: {
        "internal": True,
        "type": dict,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    FileFields.crop_state: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
}

FILE_FIELDS = FileFields(fields_props=FileFieldProps).filtered_keys('pickable', True)



DISPLAY_ADS_REQUIREMENTS = [
  {
    "name": 'Rectangle (300x250)',
    "id": 'MediumRectangle',
    "isSpecial": True,
    "img": 'https://dummyimage.com/300x250/000/fff',
    "width": 300,
    "height": 250
  },
  {
    "name": 'Rectangle Large (336x280)',
    "id": 'LargeRectangle',
    "isSpecial": True,
    "img": 'https://dummyimage.com/336x280/000/fff',
    "width": 336,
    "height": 280
  },
  {
    "name": 'Carré (250x250)',
    "width": 250,
    "height": 250,
    "id": 'Square',
    "isSpecial": True,
    "img": 'https://dummyimage.com/250x250/000/fff'
  },
  {
    "name": 'Vertical Medium (160x600)',
    "width": 160,
    "height": 600,
    "id": 'Wideskyscraper',
    "isSpecial": True,
    "img": 'https://dummyimage.com/160x600/000/fff'
  },
  {
    "name": 'Demi page (300x600)',
    "width": 300,
    "height": 600,
    "id": 'LargeSkyscraper',
    "isSpecial": True,
    "img": 'https://dummyimage.com/300x600/000/fff'
  },
  {
    "name": 'Horizontal Medium (728x90)',
    "width": 728,
    "height": 90,
    "id": 'Leaderboard',
    "isSpecial": True,
    "img": 'https://dummyimage.com/728x90/000/fff'
  },
  {
    "name": 'Horizontal (468x60)',
    "width": 468,
    "height": 60,
    "id": 'Banner',
    "isSpecial": False,
    "img": 'https://dummyimage.com/468x60/000/fff'
  },
  {
    "name": 'Vertical (120x600)',
    "width": 120,
    "height": 600,
    "id": 'Skyscraper',
    "isSpecial": False,
    "img": 'https://dummyimage.com/120x600/000/fff'
  },
  {
    "name": 'Rectangle Vertical (240x400)',
    "width": 240,
    "height": 400,
    "id": 'RV',
    "isSpecial": False,
    "img": 'https://dummyimage.com/120x600/000/fff'
  },
  {
    "name": 'Horizontal Large (970x90)',
    "width": 970,
    "height": 90,
    "id": 'LargerLeaderboard',
    "isSpecial": False,
    "img": 'https://dummyimage.com/970x90/000/fff'
  },
  {
    "name": 'Big Panneau (970x250)',
    "width": 970,
    "height": 250,
    "id": 'BigPanneau',
    "isSpecial": False,
    "img": 'https://dummyimage.com/970x250/000/fff'
  },
  {
    "name": 'Petit carré (200x200)',
    "width": 200,
    "height": 200,
    "id": 'Smallsquare',
    "isSpecial": False,
    "img": 'https://dummyimage.com/200x100/000/fff'
  },
  {
    "name": 'Not allowed (1280x720)',
    "width": 1280,
    "height": 720,
    "id": 'Smallsquare',
    "isSpecial": False,
    "img": 'https://dummyimage.com/1280x720/000/fff'
  }
];




from .workspace import (
    WorkflowWorkspace,
    WorkflowWorkspaceFields,
    WORKSPACE_COLLECTION,
    WORKSPACE_PICKABLE_FIELDS,
    WORKSPACE_MUTABLE_FIELDS,
    WORKSPACE_STATUS,
)
from .scenario import (
    WorkflowScenario,
    WorkflowScenarioFields,
    SCENARIO_COLLECTION,
    SCENARIO_PICKABLE_FIELDS,
    SCENARIO_MUTABLE_FIELDS,
    SCENARIO_STATUS,
)
from .creative import (
    WorkflowCreative,
    WorkflowCreativeFields,
    CREATIVE_COLLECTION,
    CREATIVE_PICKABLE_FIELDS,
    CREATIVE_MUTABLE_FIELDS,
    CREATIVE_TYPES,
    CREATIVE_STATUS,
)
from .discussion import (
    WorkflowDiscussionMessage,
    DISCUSSION_COLLECTION,
    DISCUSSION_PICKABLE_FIELDS,
    DISCUSSION_TYPES,
)
from .channel import (
    WorkflowChannel,
    WorkflowChannelFields,
    CHANNEL_COLLECTION,
    CHANNEL_PICKABLE_FIELDS,
    CHANNEL_MUTABLE_FIELDS,
    CHANNEL_TYPES,
)
from .channel_message import (
    WorkflowChannelMessage,
    MESSAGE_COLLECTION,
    MESSAGE_PICKABLE_FIELDS,
    MESSAGE_MUTABLE_FIELDS,
)
from .document import (
    WorkflowDocument,
    WorkflowDocumentFields,
    DOCUMENT_COLLECTION,
    DOCUMENT_PICKABLE_FIELDS,
    DOCUMENT_MUTABLE_FIELDS,
)
from .member import (
    WorkflowMember,
    WorkflowMemberFields,
    MEMBER_COLLECTION,
    MEMBER_PICKABLE_FIELDS,
    MEMBER_MUTABLE_FIELDS,
    MEMBER_ROLES,
)
from .team import (
    WorkflowTeam,
    WorkflowTeamFields,
    TEAM_COLLECTION,
    TEAM_PICKABLE_FIELDS,
    TEAM_MUTABLE_FIELDS,
)
from .invite import (
    WorkflowInvite,
    INVITE_COLLECTION,
    INVITE_PICKABLE_FIELDS,
    INVITE_STATUSES,
)
from .campaign_link import (
    WorkflowCampaignLink,
    CAMPAIGN_LINK_COLLECTION,
    CAMPAIGN_LINK_PICKABLE_FIELDS,
)



from dataclasses import dataclass
from typing import Any
from  ....utils.utils import DictUtils, BaseFieldsClass, BaseClass,init_class_kwargs,get_object_model_class

@dataclass
class WebsiteFields(BaseFieldsClass):
    id = "id"
    url = "url"
    criterionId = "criterionId"
    criterionType = "criterionType"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

@dataclass
class YoutubeChannelFields(BaseFieldsClass):
    id = "id"
    channelId = "channelId"
    name = "name"
    snippet = "snippet"
    thumbnails = "thumbnails"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

@dataclass
class YoutubeVideoFields(BaseFieldsClass):
    id = "id"
    videoId = "videoId"
    name = "name"
    snippet = "snippet"
    thumbnails = "thumbnails"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

@dataclass(init=False)
class Website(BaseClass):
    id: str
    url: str
    criterionId: int
    criterionType: str
    def __init__(self, website=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, website, None, None, None, [], **kwargs)
        for key in keys:
            setattr(self, key, cls_object[key]) 
    def to_dict(self, field=None):
        return DictUtils.remove_none_values({
            "id": self.id,
            "url": self.url,
            "criterionId": self.criterionId,
            "criterionType": self.criterionType
        })
    @staticmethod
    def from_dict(obj: Any) -> 'Website':
        cls_object, keys = get_object_model_class(obj, Website, None);

        return Website(cls_object)
@dataclass(init=False)
class YoutubeChannel(BaseClass):
    id: str
    channelId: str
    name: str
    snippet: str
    thumbnails: str
    def __init__(self, channel=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, channel, None, None, None, [], **kwargs)
        for key in keys:
            setattr(self, key, cls_object[key]) 
    def to_dict(self, field=None):
        return DictUtils.remove_none_values({
            "id": self.id,
            "channelId": self.channelId,
            "name": self.name,
            "snippet": self.snippet,
            "thumbnails": self.thumbnails
        })
    @staticmethod
    def from_dict(obj: Any) -> 'YoutubeChannel':
        cls_object, keys = get_object_model_class(obj, YoutubeChannel, None);
        return YoutubeChannel(cls_object)

@dataclass(init=False)
class YoutubeVideo(BaseClass):
    id: str
    videoId: str
    name: str
    snippet: str
    thumbnails: str
    def __init__(self, video=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, video, None, None, None, [], **kwargs)
        for key in keys:
            setattr(self, key, cls_object[key]) 
    def to_dict(self, field=None):
        return DictUtils.remove_none_values({
            "id": self.id,
            "videoId": self.videoId,
            "name": self.name,
            "snippet": self.snippet,
            "thumbnails": self.thumbnails
        })
    @staticmethod
    def from_dict(obj: Any) -> 'YoutubeVideo':
        cls_object, keys = get_object_model_class(obj, YoutubeVideo, None);
        return YoutubeVideo(cls_object)


import uuid
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Optional

from adafri.v1.base.firebase_collection import FirebaseCollectionBase

MESSAGE_COLLECTION = 'workflow_channel_messages'

MESSAGE_PICKABLE_FIELDS = [
    'id', 'channelId', 'workspaceId', 'authorUid', 'authorName',
    'content', 'threadParentId', 'reactions', 'isEdited', 'editedAt', 'createdAt',
]

MESSAGE_MUTABLE_FIELDS = ['content', 'reactions', 'isEdited', 'editedAt']


def _now() -> str:
    return datetime.now(timezone.utc).isoformat()


def _new_id() -> str:
    return str(uuid.uuid4())


@dataclass(init=False)
class WorkflowChannelMessage(FirebaseCollectionBase):
    id: str
    channelId: str
    workspaceId: str
    authorUid: str
    authorName: str
    content: str
    threadParentId: Optional[str]
    reactions: dict
    isEdited: bool
    editedAt: Optional[str]
    createdAt: str

    def __init__(self, data=None, **kwargs):
        super().__init__(collection_name=MESSAGE_COLLECTION, **kwargs)
        d = data or {}
        self.id = d.get('id') or _new_id()
        self.channelId = d.get('channelId', '')
        self.workspaceId = d.get('workspaceId', '')
        self.authorUid = d.get('authorUid', '')
        self.authorName = d.get('authorName', '')
        self.content = d.get('content', '')
        self.threadParentId = d.get('threadParentId')
        self.reactions = dict(d.get('reactions') or {})
        self.isEdited = bool(d.get('isEdited', False))
        self.editedAt = d.get('editedAt')
        self.createdAt = d.get('createdAt') or _now()

    @classmethod
    def from_dict(cls, obj: Any, **kwargs) -> 'WorkflowChannelMessage':
        return cls(obj, **kwargs)

    def to_dict(self) -> dict:
        return {
            'id': self.id,
            'channelId': self.channelId,
            'workspaceId': self.workspaceId,
            'authorUid': self.authorUid,
            'authorName': self.authorName,
            'content': self.content,
            'threadParentId': self.threadParentId,
            'reactions': self.reactions,
            'isEdited': self.isEdited,
            'editedAt': self.editedAt,
            'createdAt': self.createdAt,
        }

    @staticmethod
    def toListDict(items: list, fields=None) -> list:
        return [item.to_dict() for item in items]

    @classmethod
    def find(cls, message_id: str) -> 'Optional[WorkflowChannelMessage]':
        inst = cls()
        doc = inst.collection().document(message_id).get()
        if not doc.exists:
            return None
        return cls({**doc.to_dict(), 'id': doc.id})

    @classmethod
    def listByChannel(
        cls, channel_id: str, since: Optional[str] = None
    ) -> 'list[WorkflowChannelMessage]':
        inst = cls()
        query = (
            inst.collection()
            .where('channelId', '==', channel_id)
            .where('threadParentId', '==', None)
            .order_by('createdAt')
        )
        if since:
            query = query.where('createdAt', '>', since)
        docs = query.stream()
        return [cls({**d.to_dict(), 'id': d.id}) for d in docs]

    @classmethod
    def listThread(cls, parent_id: str) -> 'list[WorkflowChannelMessage]':
        inst = cls()
        docs = (
            inst.collection()
            .where('threadParentId', '==', parent_id)
            .order_by('createdAt')
            .stream()
        )
        return [cls({**d.to_dict(), 'id': d.id}) for d in docs]

    def save(self) -> 'WorkflowChannelMessage':
        self.collection().document(self.id).set(self.to_dict())
        return self

    def update(self, data: dict) -> 'WorkflowChannelMessage':
        filtered = {k: v for k, v in data.items() if k in MESSAGE_MUTABLE_FIELDS}
        self.collection().document(self.id).update(filtered)
        for k, v in filtered.items():
            setattr(self, k, v)
        return self

    def delete(self) -> None:
        self.collection().document(self.id).delete()



import json
import os
from dataclasses import dataclass

import pydash
from adafri.utils.utils import BaseClass, DictUtils, JsonEncoder, get_object_model_class
from firebase_admin.firestore import firestore
from google.cloud.firestore_v1.base_query import And, FieldFilter
from google.cloud.firestore_v1.batch import WriteBatch
from google.cloud.firestore_v1.field_path import FieldPath

from ...utils import ApiResponse, Error, ResponseStatus, StatusCode
from ..auth.firebase_auth import FirestoreApp


def getTimestamp():
    return firestore.SERVER_TIMESTAMP;


@dataclass(init=False)
class FirebaseCollectionBase(BaseClass):
    # _db: firestore.Client = None;
    # _fields_props = None;
    # _collection_name = None;

    def __init__(self, collection_name=None, db: firestore.Client=None, fields=None, fields_props=None, id=None, **kwargs):
        self.collection_name = collection_name;
        self.db = db;
        self.id = id;
        if self.db is None:
            self.db = FirestoreApp().firestore_client()
        super().__init__(fields=fields, fields_props=fields_props);
        
    def print_db(self):
        print('db prtiny',self.db)

    def setFields(self, fields):
        self.fields_props = fields;
    
    def generate_model(self, _key_):
        user = {};
        props = self.fields_props
        for k in DictUtils.get_keys(props):
            user[k] = props[k][_key_];
        return user;

    def collection(self, name=None):
        if name is not None:
            return self.db.collection(name)
        return self.db.collection(self.collection_name)

    def document_reference(self, _id=None) -> 'firestore.DocumentReference':
        id = _id;
        if id is None:
            if self.id is not None:
                id = self.id;
        return self.collection().document(id);

    # def get(self, id):
    #     if id is None or bool(id) is False:
    #          return None;
    #     doc = self.document_reference(id).get();
    #     if doc.exists is False:
    #         return None;
    #     return doc.to_dict();


    
    def update(self, id, data, merge=True):
        if id is None or bool(id) is False:
            return None;
        doc = self.document_reference(id);
        if doc.get().exists is False:
            return None;
        try:
            doc.set(data, merge=merge);
            return data;
        except:
            return None
    
    def remove(self, id):
        if id is None or bool(id) is False:
             return None;
        doc = self.document_reference(id);
        if doc.get().exists is False:
            return None;
        try:
            doc.delete();
            return id;
        except:
            return None

    def batchSingle(self, data_filter, batch_size = 10):
        results = []
        # Split the user_ids list into batches of 10
        for i in range(0, len(data_filter), batch_size):
            batch_ids = data_filter[i:i + batch_size]
            # Query the "users" collection with `in` filter
            query = self.collection().where(FieldPath.document_id(), "in", batch_ids)
            
            # Fetch and add results
            docs = query.stream()
            results.extend([{"id": doc.id, **doc.to_dict()} for doc in docs])
        return results

    def custom_query(self, query_params, first=True, limit: int=None, collection_name=None):
        i = 0;
        dynamic_query = None;
        filters = []
        keys = []
        values = []
        if bool(query_params) is True:
            while i < len(query_params):
                query = query_params[i];
                
                key = None;
                if 'key' in query:
                    key = query['key'];
                if key not in keys:
                    keys.append(key)
                
                comparator = None;
                if 'comp' in query:
                    comparator = query['comp']

                value = None;
                if 'value' in query:
                    value = query['value'];
                    if value not in values:
                        values.append(value)
                if None not in [key, comparator, value]:
                    filters.append(FieldFilter(key, comparator, value))
                i+=1;
        if len(keys) == 1 and len(values) > 0:
            filters = [FieldFilter(keys[0], 'in', values)]

        dynamic_query = None;
        if bool(filters) is True:
            and_filter = And(filters=filters)
            dynamic_query = self.collection(collection_name).where(filter=and_filter)
                
        else:
            dynamic_query = self.collection(collection_name)
            
        if dynamic_query is None:
            return None
        _limit = limit
        if _limit is not None and _limit > 0 or first:
            if _limit is None:
                _limit = 1
            # print('limit', _limit)
            dynamic_query = dynamic_query.limit(_limit)
        
        data = []
        values = None;
        try:
            values = dynamic_query.get();
        except Exception as query_exception:
            print('query exception', query_exception)
            return None;
        for stream in values:
            data_stream = stream.to_dict();
            # print('data stream', data_stream)
            if 'id' in data_stream and bool(data_stream['id']) is False:
                data_stream['id'] = stream.id
            data.append({"id": stream.id, **data_stream});
        if first is False:
            if len(data) == 0:
                return []
            return data;
        else:
            if len(data) == 0:
                return None;
            return data[0]

    def queryBatch(self, cls, key, data_filter, first=False):
        result = [];
        batch_size = 10
        query_result = self.batchSingle(data_filter, batch_size)
        dict_values = {"db": self.db, "collection_name": self.collection_name}
        if bool(query_result):
            if first:
                return cls.from_dict(**{key: query_result[0], **dict_values})
            else:
                for doc in query_result:
                    result.append(cls.from_dict(**{key: doc, **dict_values}))
                return result
        return [];
    
    def write_batch(self, data, type: str=None, mark_as_deleted=True, batch_size = 10):
        errors = []
        delete = []
        write = []
        if bool(data) is False or len(data)==0:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_204, None, Error(f"No data","INVALID_REQUEST", 1));

        batch = firestore.WriteBatch(self.db);
        chunk_data = pydash.chunk(data, batch_size)
        for chunk in chunk_data:
            for element in chunk:
                try:
                    if getattr(element, 'document_reference', None) is None:
                        continue;
                    if type == 'remove': 
                        if mark_as_deleted is False:
                            batch.delete(element.document_reference())
                        else:
                            batch.set(element.document_reference(), {"is_removed": True}, True)
                        delete.append(element)
                    else:
                        data_value = element.to_dict()
                        if 'createdAt' in data_value:
                            data_value['createdAt'] = firestore.SERVER_TIMESTAMP
                        batch.set(element.document_reference(), data_value, True)
                        write.append(element)
                except:
                    errors.append({"message": f"Cannot process document {element.id}"});
                    continue
            batch.commit();

        if bool(write) is False and bool(delete) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"{json.dumps(errors)}","INVALID_REQUEST", 1));
        if bool(write) is True:
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, write, None);
        if bool(delete) is True:
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, delete, None);
        return ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error(f"Something went wrong","INVALID_REQUEST", 1));

    def query(self, cls, key, query_params: list, first=False, limit=None, collection_name=None):
        result = [];
        query_result = self.custom_query(query_params, first=first, limit=limit, collection_name=collection_name)
        collection = collection_name
        if collection is None:
            collection = self.collection_name
        dict_values = {"db": self.db, "collection_name": collection}
        if bool(query_result):
            if first:
                return cls.from_dict(**{key: query_result, **dict_values})
            else:
                for doc in query_result:
                    result.append(cls.from_dict(**{key: doc, **dict_values}))
                return result
        if first:
                return None
        return [];
        
    
    # def to_json(self, _fields=None):
    #     # Convert the class attributes to a dictionary and then to JSON
    #     return json.loads(json.dumps(self, cls=JsonEncoder))
    





from dataclasses import dataclass 
from  ......utils.utils import  BaseFieldsClass
import os
from google.cloud.firestore_v1.document import Timestamp
from google.api_core.datetime_helpers import DatetimeWithNanoseconds

DISPLAY_ADS_COLLECTION = os.environ.get('GADS_ADS_DISPLAY_COLLECTION');

@dataclass
class BaseAdFields(BaseFieldsClass):
    id = "id"
    ad_group_id = "ad_group_id"
    ad_id = "ad_id"
    name = "name"
    owner = "owner"
    status = "status"
    accountId = "accountId"
    clientCustomerId = "clientCustomerId"
    createdAt = "createdAt"
    createdBy = "createdBy"
    is_removed = "is_removed"
    adgroup_id_parent = "adgroup_id_parent"
    api_version = "api_version"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

BaseAdFieldsProps = {
    BaseAdFields.id: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    BaseAdFields.name: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseAdFields.ad_id: {
        "internal": True,
        "type": int,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseAdFields.ad_group_id: {
        "internal": True,
        "type": int,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseAdFields.status: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseAdFields.adgroup_id_parent: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseAdFields.accountId: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseAdFields.clientCustomerId: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseAdFields.is_removed: {
        "internal": True,
        "type": bool,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": False,
        "default_value": False
    },
    BaseAdFields.createdBy: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": False,
        "default_value": None
    },
    BaseAdFields.createdAt: {
        "type": [int, str, Timestamp, DatetimeWithNanoseconds],
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    BaseAdFields.api_version: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": "v0"
    }
}

STANDARD_FIELDS = BaseAdFields(fields_props=BaseAdFieldsProps).filtered_keys('pickable', True)
BASE_AD_PICKABLE_FIELDS = BaseAdFields(fields_props=BaseAdFieldsProps).filtered_keys('pickable', True)



import random
import string
import uuid
from dataclasses import dataclass
from typing import Any

import pydash
from adafri.v1 import version
from firebase_admin.firestore import firestore

from ......utils import (
    ArrayUtils,
    BaseClass,
    DictUtils,
    RequestFields,
    get_object_model_class,
    get_request_fields,
    init_class_kwargs,
)
from ......utils.response import ApiResponse, Error, ResponseStatus, StatusCode
from ....models.adgroups.ads.base import BaseAd, BaseAdFieldsProps
from ....models.text import AssetText
from ....models.videos import VideoAsset
from ...models.asset import Asset as AssetNative
from .ad_fields import (
    RESPONSIVE_DISPLAY_AD_PICKABLE_FIELDS,
    RESPONSIVE_DISPLAY_ADS_COLLECTION,
    ResponsiveDisplayAdFields,
    ResponsiveDisplayAdFieldsProps,
)


def get_usage(key: str):
    if key == 'marketingImages':
        return ["image"], "rect"
    if key == 'squareMarketingImages':
        return ["image"], "sqr"
    if key=='squareLogoImages':
        return ["logo"], "sqr"
    if key in ['landscapeLogoImages', 'logoImages']:
        return ["logo"], "rect"
    return None

def filter_responsive_display_ad_request_fields(fields, default_fields = RESPONSIVE_DISPLAY_AD_PICKABLE_FIELDS):
    request_fields = get_request_fields(fields, default_fields, [default_fields[0]])
    if request_fields is None or bool(request_fields) is False:
        request_fields = [default_fields[0]]
    return request_fields

@dataclass(init=False)
class NativeAdsToPublish(BaseClass):
    id: str
    titles: list[AssetText]
    descriptions: list[AssetText]
    longHeadline: str
    brand: AssetText
    videosAssets: list[VideoAsset]
    marketingImages: list[AssetNative]
    squareMarketingImages: list[AssetNative]
    logoImages: list[AssetNative]
    landscapeLogoImages: list[AssetNative]
    is_removed: bool

    def normalize(self, images: list[AssetNative], key: str):
        if bool(images) is False:
            return []
        u = get_usage(key)
        if u is None:
            return []
        # print('normalize', u)
        # print('images', len(images))
        usage, useFor = u
        for im in images:
            if bool(im.usage) is False:
                im.usage = usage
            if bool(im.useFor) is False:
                im.useFor = useFor
            if bool(im.fullSizeInfo) and bool(im.fullSizeInfo.imageUrl):
                if bool(im.imageUrl) is False:
                    im.imageUrl = im.fullSizeInfo.imageUrl
                    im.file_url = im.fullSizeInfo.imageUrl
                if bool(im.width) is False:
                    im.width = im.fullSizeInfo.imageWidth
                if bool(im.height) is False:
                    im.height = im.fullSizeInfo.imageHeight
            if bool(im.imageFileSize) and bool(im.size) is False:
                im.size = im.imageFileSize
        return images
    def __init__(self, asset=None, **kwargs):
        super().__init__(**kwargs)
        (cls_object, keys, data_args) = init_class_kwargs(self, asset, None, None, None, [], **kwargs)
        for key in keys:
            if key=='titles':
                setattr(self, key, AssetText().fromListDict(cls_object[key]))
            elif key=='descriptions':
                setattr(self, key, AssetText().fromListDict(cls_object[key]))
            elif key=='brand':
                setattr(self, key, AssetText().from_dict(cls_object[key]))
            elif key=='videosAssets':
                setattr(self, key, VideoAsset().fromListDict(cls_object[key]))
            elif key in ['marketingImages', 'squareMarketingImages', 'logoImages', 'landscapeLogoImages']:
                images = AssetNative().fromListDict(cls_object[key])
                setattr(self, key,self.normalize(images, key))
            else:
                setattr(self, key, cls_object[key]) 
    def to_dict(self, field=None):
        return DictUtils.remove_none_values({
            "id": self.id,
            "titles": AssetText().toListDict(self.titles, None),
            "descriptions": AssetText().toListDict(self.descriptions, None),
            "longHeadline": self.longHeadline,
            "brand": self.brand.to_dict(None),
            "videosAssets": VideoAsset().toListDict(self.videosAssets, None),
            "marketingImages": AssetNative().toListDict(self.marketingImages, None),
            "squareMarketingImages": AssetNative().toListDict(self.squareMarketingImages, None),
            "logoImages": AssetNative().toListDict(self.logoImages, None),
            "landscapeLogoImages": AssetNative().toListDict(self.landscapeLogoImages, None),
            "is_removed": self.is_removed
        })
    @staticmethod
    def from_dict(obj: Any) -> 'NativeAdsToPublish':
        cls_object, keys = get_object_model_class(obj, NativeAdsToPublish, None);
        return NativeAdsToPublish(cls_object)
    @staticmethod
    def convert_to_responsive_display_ad(value):
        ad = ResponsiveDisplayAd();
        if bool(value.titles) is True:
            ad.headlines = value.titles.copy();
        if bool(value.descriptions) is True:
            ad.descriptions = value.descriptions;
        if bool(value.longHeadline) is True:
            ad.longHeadline = value.longHeadline;
        if bool(value.brand) is True:
            ad.businessName = value.brand;
        if bool(value.videosAssets) is True:
            ad.youtubeVideos = value.videosAssets;
        if bool(value.marketingImages) is True:
            ad.marketingImages = value.marketingImages.copy();
        if bool(value.squareMarketingImages) is True:
            ad.squareMarketingImages = value.squareMarketingImages.copy();
        if bool(value.logoImages) is True:
            ad.squareLogoImages = value.logoImages.copy();
        if bool(value.landscapeLogoImages) is True:
            ad.logoImages = value.landscapeLogoImages.copy();
        if getattr(value, 'owner', None) is not None:
            ad.owner = value.owner
        ad.is_published = False
        ad.adData = NativeAdsToPublish.from_dict(value.to_dict());
        return ad

    def to_responsive_display_ad(self, owner=None):
        if bool(owner) is True:
            self.owner = owner
        v = NativeAdsToPublish.convert_to_responsive_display_ad(self);
        return v
@dataclass(init=False)
class ResponsiveDisplayAd(BaseAd):
    headlines: list[AssetText]
    descriptions: list[AssetText]
    longHeadline: AssetText
    businessName: str
    marketingImages: list[AssetNative]
    squareMarketingImages: list[AssetNative]
    logoImages: list[AssetNative]
    squareLogoImages: list[AssetNative]
    youtubeVideos: list[VideoAsset]
    resourceName: str
    adData: NativeAdsToPublish
    displayUrl: str
    finalAppUrls: list[str]
    finalMobileUrls: list[str]
    finalUrls: list[str]
    responsive_display_ad: dict
    is_published: bool = False
    adgroup: Any | None = None
    __baseFields = ResponsiveDisplayAdFieldsProps
    class_object = None;

    def normalize(self, images: list[AssetNative], key: str):
        if bool(images) is False:
            return []
        u = get_usage(key)
        if u is None:
            return []
        # print('normalize', u)
        # print('images', len(images))
        usage, useFor = u
        # print('usage', usage, useFor)
        for im in images:
            # print('im.usage', im.usage)
            if bool(im.usage) is False:
                im.usage = usage
            # if bool(im.useFor) is False:
            im.useFor = useFor
            if bool(im.fullSizeInfo) and bool(im.fullSizeInfo.imageUrl):
                if bool(im.imageUrl) is False:
                    im.imageUrl = im.fullSizeInfo.imageUrl
                if bool(im.file_url) is False:
                    im.file_url = im.fullSizeInfo.imageUrl
                if bool(im.url) is False:
                    # print('im.url', im.url)
                    im.url = im.fullSizeInfo.imageUrl
                if bool(im.width) is False:
                    im.width = im.fullSizeInfo.imageWidth
                if bool(im.height) is False:
                    im.height = im.fullSizeInfo.imageHeight
            if bool(im.imageFileSize) and bool(im.size) is False:
                im.size = im.imageFileSize
        return images
    def __init__(self, ad=None, **kwargs):
        _ad = ad
        if isinstance(ad, str):
            _ad = {"id": ad}
        (cls_object, keys, data_args) = init_class_kwargs(self, _ad, RESPONSIVE_DISPLAY_AD_PICKABLE_FIELDS, ResponsiveDisplayAdFieldsProps, RESPONSIVE_DISPLAY_ADS_COLLECTION, ['id'], **kwargs)
        super().__init__(**{**data_args, "ad": _ad});
        for key in keys:
            if BaseAdFieldsProps.get(key) is None:
                if key == "adData":
                    setattr(self, key, NativeAdsToPublish().from_dict(cls_object[key]))
                elif key == 'headlines':
                    setattr(self, key, AssetText().fromListDict(cls_object[key]))
                elif key == 'descriptions':
                    setattr(self, key, AssetText().fromListDict(cls_object[key]))
                elif key in ['marketingImages', 'squareMarketingImages', 'logoImages', 'squareLogoImages']:
                    setattr(self, key, self.normalize(AssetNative().fromListDict(cls_object[key]), key))
                elif key == "longHeadline":
                    
                    if isinstance(cls_object[key], str):
                        setattr(self, key, AssetText.from_dict({'assetText': cls_object[key]}))
                    elif isinstance(cls_object[key], dict) is True and 'assetText' in cls_object[key]:
                        setattr(self, key, AssetText.from_dict(cls_object[key]))
                    else:
                        setattr(self, key, AssetText.from_dict({'assetText': None}))
                elif key == "businessName":
                    if isinstance(cls_object[key], str):
                        # setattr(self, key, AssetText.from_dict({'assetText': cls_object[key]}))
                        setattr(self, key, cls_object[key])
                    elif isinstance(cls_object[key], dict) is True and 'assetText' in cls_object[key]:
                        setattr(self, key, cls_object[key]['assetText'])
                        # setattr(self, key, AssetText.from_dict(cls_object[key]))
                    else:
                        setattr(self, key, None)
                elif key == 'youtubeVideos':
                    setattr(self, key, VideoAsset().fromListDict(cls_object[key]))
                else:
                    setattr(self, key, cls_object[key])
        if self.ad_id is None or self.ad_id == 0:
            self.is_published = False
        else:
            self.is_published = True
        self.class_object = ResponsiveDisplayAd


    @staticmethod
    def generate_model(_key_="default_value"):
        model = {};
        props = ResponsiveDisplayAdFieldsProps
        for k in DictUtils.get_keys(props):
            if k == "api_version":
                model[k] = version
                continue;
            if _key_ in props[k] and props[k][_key_] is not None:
                model[k] = props[k][_key_];
        
        # Generate random campaign name
        random_suffix = ''.join(random.choices(string.ascii_uppercase + string.digits, k=6))
        model['name'] = f'Ad Display {random_suffix}'
        return model;

    @staticmethod
    def from_dict(ad: Any, db=None, collection_name=None) -> 'ResponsiveDisplayAd':
        if isinstance(ad, dict) is True:
            if 'squareMarketingImage' in ad:
                ad['squareMarketingImages'] = ad['squareMarketingImage']
            if 'brand' in ad and 'businessName' not in ad:
                ad['businessName'] = ad['brand']
            if 'longHeadline' in ad and type(ad['longHeadline'])is str:
                ad['longHeadline'] = {"assetText": ad['longHeadline']} 
        cls_object, keys = get_object_model_class(ad, ResponsiveDisplayAd, ResponsiveDisplayAdFieldsProps);
        _ad = ResponsiveDisplayAd(cls_object, db=db, collection_name=collection_name)
        return _ad
    def generateAd(self, adgroup, data=None, only_generate = False):
        if adgroup is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Adgroup not found","INVALID_REQUEST", 1))

        model = ResponsiveDisplayAd.generate_model()
        model['accountId'] = adgroup.accountId
        model['clientCustomerId'] = adgroup.clientCustomerId
        if data is not None:
            model = DictUtils.merge_dict(data, model, True)
        # print('create ad with model', model)
        create = ResponsiveDisplayAd().create(model, adgroup, only_generate)
        if create.status == 'error':
            return create
        # print('fields', CAMPAIGN_DISPLAY_PICKABLE_FIELDS)
        values = create.data
        # values = DictUtils.omit_fields(values, RESPONSIVE_DISPLAY_AD_PICKABLE_FIELDS + ['owner', 'id_campagne', 'ad_group_id'])
        # print('values', values)
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, values, None)
    
    @staticmethod
    def property_keys():
        keys = [
                ResponsiveDisplayAdFields.marketingImages, 
                ResponsiveDisplayAdFields.squareMarketingImages,
                ResponsiveDisplayAdFields.logoImages,
                ResponsiveDisplayAdFields.squareLogoImages, 
                ResponsiveDisplayAdFields.name, 
                ResponsiveDisplayAdFields.adData,
                ResponsiveDisplayAdFields.headlines,
                ResponsiveDisplayAdFields.descriptions,
                ResponsiveDisplayAdFields.longHeadline,
                ResponsiveDisplayAdFields.businessName,
                ResponsiveDisplayAdFields.displayUrl,
                ResponsiveDisplayAdFields.finalUrls,
                ResponsiveDisplayAdFields.finalMobileUrls,
                ResponsiveDisplayAdFields.finalAppUrls]
        return keys
    @staticmethod
    def get_ad_group_props(value):
        if value is None:
            return {}
        return DictUtils.pick_fields(value, ResponsiveDisplayAd.property_keys())
    def to_dict(self, fields=None):
        base = super().to_dict(fields)
        adData = None;
        if self.adData is not None:
            adData = self.adData.to_dict()
        return DictUtils.remove_none_values({
            **base,
            "headlines": AssetText().toListDict(self.headlines, None),
            "descriptions": AssetText().toListDict(self.descriptions, None),
            "longHeadline": self.longHeadline.to_dict(),
            "businessName": self.businessName,
            "marketingImages": AssetNative().toListDict(self.marketingImages, None),
            "squareMarketingImages": AssetNative().toListDict(self.squareMarketingImages, None),
            "logoImages": AssetNative().toListDict(self.logoImages, None),
            "squareLogoImages": AssetNative().toListDict(self.squareLogoImages, None),
            "youtubeVideos": VideoAsset().toListDict(self.youtubeVideos, None),
            "adData": adData,
            "responsive_display_ad": self.responsive_display_ad
    })
    @staticmethod
    def findAd(id, aacid=None):
        print('findAd', id, aacid)
        isID = False
        try:
            id = int(id)
        except Exception as e:
            isID = True
        if isID is True:
            ad = ResponsiveDisplayAd.from_dict({"id": id}).get(RESPONSIVE_DISPLAY_AD_PICKABLE_FIELDS)
            if ad is None:
                return None
            return ad 
        if id is None or aacid is None or not isinstance(id, int) or not isinstance(aacid, str) or len(aacid) == 0:
            return None
        ad = ResponsiveDisplayAd(None).query(ResponsiveDisplayAd, "ad", query_params=[{"key": "accountId", "comp": "==", "value": aacid}, {"key": "ad_group_id", "comp": "==", "value": id}], first=True)
        return ad
    def getAds(self, adgroup, fields=['id', 'name'], response_type = 'dict'):
        isID = False
        if adgroup is None:
            return None
        adGroupId = adgroup.id
        # print('adgroup version', adgroup.api_version)
        if adgroup.api_version == 'v0' and getattr(adgroup, 'ad_group_id', None) is not None:
            adGroupId = adgroup.ad_group_id
        try:
            adGroupId = int(adGroupId)
        except:
            isID = True
        if isID is False and adgroup.accountId is None:
            return None
        # print('adGroupId', adGroupId)
        # print('isID', isID)
        ads = []
        queryParams = [{"key": "accountId", "comp": "==", "value": adgroup.accountId}]
        if isID is False:
            queryParams.append({"key": "ad_group_id", "comp": "==", "value": adGroupId})
        else :
            queryParams.append({"key": "adgroup_id_parent", "comp": "==", "value": adGroupId})
        ads_ = self.query(ResponsiveDisplayAd, "ad", query_params=queryParams)
        
        request_fields = filter_responsive_display_ad_request_fields(fields)
        # print('request_fields', request_fields)
        for ad in ads_:
            if response_type == 'class':
                ads.append(ad)
            else:
                if fields is None:
                    ads.append(ad.to_dict())
                    continue
                # ads.append(DictUtils.filter_by_fields(ad.to_dict(), request_fields))
                ads.append(ad.to_dict())
        return ads
    
    def get(self, fields=None):
        if bool(self.id) is None:
            return None;
        request_fields = filter_responsive_display_ad_request_fields(fields)
        doc = self.document_reference().get();
        if doc.exists is False:
            return None;
        doc_dict = doc.to_dict()
        if 'id' in doc_dict and bool(doc_dict['id']) is False:
            del doc_dict['id']
        data = {"id": doc.id, **doc_dict}
        return ResponsiveDisplayAd.from_dict(ad=ResponsiveDisplayAd.from_dict(ad=data).to_dict(),  db=self.db, collection_name=self.collection_name);

    def getByUserId(self,id, fields=None):
        ads = [];
        request_fields = filter_responsive_display_ad_request_fields(fields)
        docs = self.collection().where('owner','==',id).get();
        for doc in docs:
            data = doc.to_dict();
            data['id'] = doc.id;
            ads.append(ResponsiveDisplayAd.from_dict(ad=ResponsiveDisplayAd.from_dict(ad=data).to_json(request_fields),  db=self.db, collection_name=self.collection_name));
        return ads;
    def normalize_response(self):
        data = ResponsiveDisplayAd.from_dict(DictUtils.pick_fields(self.to_dict(), RESPONSIVE_DISPLAY_AD_PICKABLE_FIELDS))
        return ResponsiveDisplayAd.from_dict(DictUtils.omit_fields(data.to_dict(), ["owner", "createdBy"]))

    def getChangedFields(self, data):
        last_value = self.to_dict();
        filtered_value = pydash.pick(data, ResponsiveDisplayAdFields().filtered_keys('editable', True));
        new_value = DictUtils.merge_dict(filtered_value, last_value, True);
        _changed_fields = DictUtils.compare_nested_objects(last_value, new_value);
        changed_fields = ArrayUtils.filter(DictUtils.get_keys(_changed_fields), lambda x: str(x).find('.') == -1)
        return changed_fields

    def build_ad_image_asset(self, data: list[AssetNative], usage: list[str], useFor: str, test = False):
        images_assets = []
        error = None
        for image_asset in data:
            image_asset.accountId = self.accountId
            image_asset.owner = self.owner
            image_asset.usage = usage.copy()
            image_asset.useFor = useFor
            image_asset.adgroup = self.adgroup
            if not bool(image_asset.clientCustomerId):
                image_asset.clientCustomerId = self.clientCustomerId
            image: ApiResponse = image_asset.build_asset(test=test);
            if image.status == ResponseStatus.OK:
                images_assets.append(image.data)
            else:
                error = image
                break;
        if error is not None and bool(error) is True:
            return error
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, images_assets, None);

    def validate_asset_text(self, values: list[AssetText], title: str, minLength = 0, maxLength = 90):
        unique: list[AssetText] = []
        error = None
        for _value in values.copy():
            value = AssetText()
            value.assetText = _value.assetText.strip()
            validate = self.validate_token_string(value.assetText, f"{title}", minLength, maxLength, ["'"])
            if validate.status == ResponseStatus.ERROR:
                error = validate
                break
            text_value = self.sanitize_text(value.assetText, maxLength)
            if True in [u.assetText==text_value for u in unique]:
                error = ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error(f"Duplicate {title} {value.assetText}", "INVALID_REQUEST", 1))
                break
            value.assetText = text_value
            unique.append(value)
        if error is not None:
            return error
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, unique, None);
    def validate_token_string(self, value: str | None, kind: str, min_length = 0, max_length = 25, chars_extended: list[str] | None = None):
        """
        Validate business name for Google Ads Responsive Display Ads
        
        Args:
            value (str): The business name to validate
            
        Returns:
            dict: Validation result with 'valid' boolean and 'message' string
        """
        MAX_LENGTH = max_length
        MIN_LENGTH = min_length
        
        # Check if business name is provided
        if not value:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error(f"{kind} is required", "INVALID_REQUEST", 1))
        
        # Check if it's a string
        if not isinstance(value, str):
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error(f"{kind} {value} must be a string", "INVALID_REQUEST", 1))
        
        # Strip whitespace
        value = value.strip()
        
        # Check if empty after stripping
        if not value:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error(f"{kind} cannot be empty", "INVALID_REQUEST", 1))
        
        # Check length
        if len(value) < MIN_LENGTH:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error(f"{kind} {value} must be at least {MIN_LENGTH} characters long", "INVALID_REQUEST", 1))
        if len(value) > MAX_LENGTH:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error(f"{kind} {value} must be {MAX_LENGTH} characters or fewer. Current length: {len(value)}", "INVALID_REQUEST", 1))
        
        # Check for invalid characters (optional - based on your requirements)
        # Google Ads typically allows letters, numbers, spaces, and common punctuation
        chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789 -.,&()'
        if chars_extended is not None:
            chars += ''.join(chars_extended)
        invalid_chars = set(value) - set(chars)
        if invalid_chars:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error(f"{kind} {value} contains invalid characters: {', '.join(invalid_chars)}", "INVALID_REQUEST", 1))
        
        # return {
        #     'valid': True,
        #     'message': 'Business name is valid',
        #     'length': len(value),
        #     'truncated_name': value[:MAX_LENGTH] if len(value) > MAX_LENGTH else value
        # }
        
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, None, None);

    def sanitize_text(self, value: str, max_length: int = 25) -> str:
        """
        Sanitize and truncate string to meet Google Ads requirements
        
        Args:
            value (str): The string to sanitize
            max_length (int): Maximum allowed length (default: 25)
            
        Returns:
            str: Sanitized string
        """
        if not value:
            return ""
        
        # Convert to string and strip whitespace
        sanitized = str(value).strip()
        
        # Remove invalid characters (keep only alphanumeric, spaces, and common punctuation)
        import re
        sanitized = re.sub(r'[^a-zA-Z0-9\s\-.,&()]', '', sanitized)
        
        # Truncate to max length
        if len(sanitized) > max_length:
            sanitized = sanitized[:max_length].rstrip()
        
        return sanitized

    def check_point_ad(self, test=False):
        if bool(self.name) is False:
            self.name = f"new Ad {uuid.uuid1()}"
        if bool(self.businessName):
            value_checkpoint = self.validate_token_string(self.businessName, "Business name")
            if value_checkpoint.status==ResponseStatus.ERROR:
                return value_checkpoint
            self.businessName = self.sanitize_text(self.businessName)
        headlines_checkpoint = self.validate_asset_text(self.headlines, "headline", 3, 30)
        if headlines_checkpoint.status==ResponseStatus.ERROR:
            return headlines_checkpoint
        self.headlines = headlines_checkpoint.data
        descriptions_checkpoint = self.validate_asset_text(self.descriptions, "description", 3, 90)
        if descriptions_checkpoint.status==ResponseStatus.ERROR:
            return descriptions_checkpoint
        self.descriptions = descriptions_checkpoint.data
        longHeadline_checkpoint = self.validate_asset_text([self.longHeadline], "longHeadline", 3, 90)
        if longHeadline_checkpoint.status==ResponseStatus.ERROR:
            return longHeadline_checkpoint
        self.longHeadline = longHeadline_checkpoint.data[0]
        marketing_images_checkpoint = self.build_ad_image_asset(self.marketingImages, ['image'], 'rect', test)
        if marketing_images_checkpoint.status == ResponseStatus.ERROR:
            return marketing_images_checkpoint
        self.marketingImages = marketing_images_checkpoint.data
        print('marketingImages', len(self.marketingImages))
        square_marketing_images_checkpoint = self.build_ad_image_asset(self.squareMarketingImages, ['image'], 'sqr', test)
        if square_marketing_images_checkpoint.status == ResponseStatus.ERROR:
            return square_marketing_images_checkpoint
        self.squareMarketingImages = square_marketing_images_checkpoint.data
        print('squareMarketingImages', len(self.squareMarketingImages))
        if bool(self.squareLogoImages):
            square_logo_images_checkpoint = self.build_ad_image_asset(self.squareLogoImages, ['logo'], 'sqr', test)
            if square_logo_images_checkpoint.status == ResponseStatus.ERROR:
                return square_logo_images_checkpoint
            self.squareLogoImages = square_logo_images_checkpoint.data
            print("sqaureLogoImages", len(self.squareLogoImages))
        if bool(self.logoImages):
            logo_images_checkpoint = self.build_ad_image_asset(self.logoImages, ['logo'], 'rect', test)
            if logo_images_checkpoint.status == ResponseStatus.ERROR:
                return logo_images_checkpoint
            self.logoImages = logo_images_checkpoint.data
            print("logoImages", len(self.logoImages))
        # print('check done ==>', self.to_dict())
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, self, None)

    def create(self, data, adgroup, only_generate = False):
        # creation_date = created_at * 1000
        # return DateUtils.from_timestamp(creation_date).isoformat()
        # try:
        # print('campaign account', data)
        self.adgroup = adgroup
        data_create = DictUtils.dict_from_keys(data, ResponsiveDisplayAdFields().filtered_keys('mutable', True))
        if bool(data_create) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Invalid request","INVALID_REQUEST", 1))
        data_create = DictUtils.merge_dict({
                "accountId": adgroup.accountId,
                "owner": adgroup.owner,
                "clientCustomerId": adgroup.clientCustomerId,
                "adgroup_id_parent": adgroup.id,
                "createdBy": f"users/{adgroup.owner}"
            }, data_create.copy(),True)
        # print('data_create', data_create)
        document_id = self.collection().document().id
        # exist = self.query(ResponsiveDisplayAd, "ad", [{"key": "name", "comp": "==", "value": data_create['name']}], True)
        # if exist is not None:
        #     return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Adgroup name {data_create['name']} already exists","INVALID_REQUEST", 1))
        model = ResponsiveDisplayAd.generate_model()
        model['id'] = document_id
        data_create = DictUtils.merge_dict(data_create, model, True)
        ad = ResponsiveDisplayAd.from_dict(dict(data_create).copy())
        ad.adgroup = adgroup
        checkpoint = ad.check_point_ad(only_generate)
        if checkpoint.status == ResponseStatus.ERROR:
            return checkpoint
        return checkpoint
        create = {**ad.to_dict(), 'createdAt': firestore.SERVER_TIMESTAMP}
        print('creating responsive search ad in few')
        if only_generate is False or only_generate is None:
            doc_ref = self.collection().document(document_id)
            doc_ref.set(create, True)
        ad = ResponsiveDisplayAd.from_dict(create).normalize_response()
        if ad is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Failed to create ad","INVALID_REQUEST", 1))
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, ad.normalize_response(), None)  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the
        # except Exception as e:
        #     print(e)
        #     return {"error": "Exception creating campaign"};

    def update(self, data):
        try:
            last_value = self.to_dict();
            print('data update', data)
            # print('last', last_value);
            filtered_value = pydash.pick(data, ResponsiveDisplayAdFields().filtered_keys('editable', True));
            # print('filtered', filtered_value);
            new_value = DictUtils.merge_dict(filtered_value, last_value, True);
            _changed_fields = DictUtils.compare_nested_objects(last_value, new_value);
            changed_fields = ArrayUtils.filter(DictUtils.get_keys(_changed_fields), lambda x: str(x).find('.') == -1)
            data_update = DictUtils.dict_from_keys(filtered_value, changed_fields);
            if bool(data_update) is False:
                return ApiResponse(ResponseStatus.OK, StatusCode.status_200, data_update, None);
            # try:
            #     self.document_reference().set(data_update, merge=True)
            # except Exception as e:
            #     return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(str(e),"INVALID_REQUEST", 1))
            # return DictUtils.dict_from_keys(self.get(), changed_fields);
            # fields = changed_fields
            print('data_update ad', data_update)
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, data_update, None);
        except Exception as e:
            print(e)
            return None;

    def to_json(self, _fields=None):
        return self.to_dict()
    def remove(self, only_mark_as_removed=True):
        if self.is_removed is True:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error(f"Adgroup {self.id} already removed","INVALID_REQUEST", 1));
        try:
            if self.id is None:
                return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot identify Campaign with id {self.id}","INVALID_REQUEST", 1));
            if only_mark_as_removed:
                self.document_reference().set({"is_removed": True}, merge=True)
            else:
                self.document_reference().delete();
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, {"message": f"Campaign {self.id} deleted"}, None);
        except:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"An error occurated while removing authorization code with id {self.id}","INVALID_REQUEST", 1));




import uuid
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Optional

from adafri.v1.base.firebase_collection import FirebaseCollectionBase

CHANNEL_COLLECTION = 'workflow_channels'

CHANNEL_PICKABLE_FIELDS = [
    'id', 'workspaceId', 'name', 'slug', 'description',
    'type', 'isDefault', 'ownerId',
    'allowedTeamIds', 'allowedMemberUids', 'pinnedMessageIds',
    'createdAt', 'createdBy',
]

CHANNEL_MUTABLE_FIELDS = [
    'name', 'slug', 'description', 'type',
    'allowedTeamIds', 'allowedMemberUids', 'pinnedMessageIds',
]

CHANNEL_TYPES = ['general', 'announcements', 'creatives', 'scenarios', 'custom']


def _now() -> str:
    return datetime.now(timezone.utc).isoformat()


def _new_id() -> str:
    return str(uuid.uuid4())


class WorkflowChannelFields:
    @staticmethod
    def check_create_fields(data: dict) -> dict:
        if not data:
            return {'error': 'Missing data'}
        if not data.get('name', '').strip():
            return {'error': 'name is required'}
        if not data.get('workspaceId', '').strip():
            return {'error': 'workspaceId is required'}
        return data

    @staticmethod
    def mutable_keys() -> list:
        return CHANNEL_MUTABLE_FIELDS


@dataclass(init=False)
class WorkflowChannel(FirebaseCollectionBase):
    id: str
    workspaceId: str
    name: str
    slug: str
    description: str
    type: str
    isDefault: bool
    ownerId: str
    allowedTeamIds: list
    allowedMemberUids: list
    pinnedMessageIds: list
    createdAt: str
    createdBy: str

    def __init__(self, data=None, **kwargs):
        super().__init__(collection_name=CHANNEL_COLLECTION, **kwargs)
        d = data or {}
        name = d.get('name', '')
        self.id = d.get('id') or _new_id()
        self.workspaceId = d.get('workspaceId', '')
        self.name = name
        self.slug = d.get('slug') or name.lower().replace(' ', '-')
        self.description = d.get('description', '')
        self.type = d.get('type', 'custom')
        self.isDefault = bool(d.get('isDefault', False))
        self.ownerId = d.get('ownerId', '')
        self.allowedTeamIds = list(d.get('allowedTeamIds') or [])
        self.allowedMemberUids = list(d.get('allowedMemberUids') or [])
        self.pinnedMessageIds = list(d.get('pinnedMessageIds') or [])
        self.createdAt = d.get('createdAt') or _now()
        self.createdBy = d.get('createdBy', '')

    @classmethod
    def from_dict(cls, obj: Any, **kwargs) -> 'WorkflowChannel':
        return cls(obj, **kwargs)

    def to_dict(self) -> dict:
        return {
            'id': self.id,
            'workspaceId': self.workspaceId,
            'name': self.name,
            'slug': self.slug,
            'description': self.description,
            'type': self.type,
            'isDefault': self.isDefault,
            'ownerId': self.ownerId,
            'allowedTeamIds': self.allowedTeamIds,
            'allowedMemberUids': self.allowedMemberUids,
            'pinnedMessageIds': self.pinnedMessageIds,
            'createdAt': self.createdAt,
            'createdBy': self.createdBy,
        }

    @staticmethod
    def toListDict(items: list, fields=None) -> list:
        return [item.to_dict() for item in items]

    @classmethod
    def find(cls, channel_id: str) -> 'Optional[WorkflowChannel]':
        inst = cls()
        doc = inst.collection().document(channel_id).get()
        if not doc.exists:
            return None
        return cls({**doc.to_dict(), 'id': doc.id})

    @classmethod
    def listByWorkspace(cls, workspace_id: str) -> 'list[WorkflowChannel]':
        inst = cls()
        docs = (
            inst.collection()
            .where('workspaceId', '==', workspace_id)
            .order_by('createdAt')
            .stream()
        )
        return [cls({**d.to_dict(), 'id': d.id}) for d in docs]

    def save(self) -> 'WorkflowChannel':
        self.collection().document(self.id).set(self.to_dict())
        return self

    def update(self, data: dict) -> 'WorkflowChannel':
        filtered = {k: v for k, v in data.items() if k in CHANNEL_MUTABLE_FIELDS}
        self.collection().document(self.id).update(filtered)
        for k, v in filtered.items():
            setattr(self, k, v)
        return self

    def delete(self) -> None:
        self.collection().document(self.id).delete()
